from typing import TYPE_CHECKING, Any, Union

import boto3
from pystac.link import Link
from pystac.stac_io import StacIO

from .boto3_config import CONFIG
from .s3_utils import get_bucket_and_key_from_url

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_s3 import S3Client
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3Client = object  # pragma: no mutate

S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)


class S3StacIO(StacIO):
    def read_text(self, source: Union[str, Link], *_args: Any, **_kwargs: Any) -> str:
        url = source.href if isinstance(source, Link) else source
        bucket, key = get_bucket_and_key_from_url(url)
        obj = S3_CLIENT.get_object(Bucket=bucket, Key=key)
        result: str = obj["Body"].read().decode("utf-8")

        return result

    def write_text(self, dest: Union[str, Link], txt: str, *_args: Any, **_kwargs: Any) -> None:
        url = dest.href if isinstance(dest, Link) else dest
        bucket, key = get_bucket_and_key_from_url(url)
        S3_CLIENT.put_object(Bucket=bucket, Key=key, Body=txt.encode())



from enum import Enum
from os import environ
from typing import Any, Dict, Optional, Tuple, Type

from pynamodb.attributes import MapAttribute, UnicodeAttribute
from pynamodb.indexes import AllProjection, GlobalSecondaryIndex
from pynamodb.models import MetaModel, Model

from .aws_keys import AWS_DEFAULT_REGION_KEY
from .check import Check
from .models import CHECK_ID_PREFIX, DB_KEY_SEPARATOR, URL_ID_PREFIX
from .parameter_store import ParameterName, get_param
from .types import JsonObject


class ValidationResult(Enum):
    FAILED = "Failed"
    PASSED = "Passed"


# TODO: Remove inherit-non-class when astroid is at version 2.6 or later pylint:disable=fixme
class ValidationOutcomeIdx(
    GlobalSecondaryIndex["ValidationResultsModelBase"]
):  # pylint:disable=too-few-public-methods,inherit-non-class
    class Meta:  # pylint:disable=too-few-public-methods

        index_name = "validation_outcome"
        read_capacity_units = 1
        write_capacity_units = 1
        projection = AllProjection()

    pk = UnicodeAttribute(hash_key=True, attr_name="pk")
    result = UnicodeAttribute(range_key=True, attr_name="result")


class ValidationResultsModelBase(Model):
    pk = UnicodeAttribute(hash_key=True)
    sk = UnicodeAttribute(range_key=True)
    result = UnicodeAttribute()
    # TODO: Remove type-arg when PynamoDB issue #920 is fixed pylint:disable=fixme
    details: MapAttribute[str, Any] = MapAttribute(null=True)  # type: ignore[no-untyped-call]

    validation_outcome_index: ValidationOutcomeIdx


def validation_results_model_with_meta(
    results_table_name: Optional[str] = None,
) -> Type[ValidationResultsModelBase]:
    if results_table_name is None:
        results_table_name = get_param(ParameterName.STORAGE_VALIDATION_RESULTS_TABLE_NAME)

    class ValidationResultsModelMeta(MetaModel):
        def __new__(
            cls,
            name: str,
            bases: Tuple[Type[object], ...],
            namespace: Dict[str, Any],
            discriminator: Optional[Any] = None,
        ) -> "ValidationResultsModelMeta":
            namespace["Meta"] = type(
                "Meta",
                (),
                {
                    "table_name": results_table_name,
                    "region": environ[AWS_DEFAULT_REGION_KEY],
                },
            )
            klass: "ValidationResultsModelMeta"
            klass = MetaModel.__new__(  # type: ignore[no-untyped-call]
                cls, name, bases, namespace, discriminator=discriminator
            )
            return klass

    class ValidationResultsModel(ValidationResultsModelBase, metaclass=ValidationResultsModelMeta):
        validation_outcome_index = ValidationOutcomeIdx()

    return ValidationResultsModel


class ValidationResultFactory:  # pylint:disable=too-few-public-methods
    def __init__(
        self,
        hash_key: str,
        results_table_name: str,
    ):

        self.hash_key = hash_key
        self.validation_results_model = validation_results_model_with_meta(results_table_name)

    def save(
        self, url: str, check: Check, result: ValidationResult, details: Optional[JsonObject] = None
    ) -> None:
        self.validation_results_model(
            pk=self.hash_key,
            sk=f"{CHECK_ID_PREFIX}{check.value}{DB_KEY_SEPARATOR}{URL_ID_PREFIX}{url}",
            result=result.value,
            details=details,
        ).save()



from json import dumps
from os.path import basename
from typing import TYPE_CHECKING, List
from urllib.parse import quote, urlparse
from uuid import uuid4

import boto3
import smart_open
from jsonschema import ValidationError, validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY
from ..boto3_config import CONFIG
from ..error_response_keys import ERROR_KEY, ERROR_MESSAGE_KEY
from ..import_dataset_keys import NEW_KEY_KEY, ORIGINAL_KEY_KEY, TARGET_BUCKET_NAME_KEY
from ..import_file_batch_job_id_keys import ASSET_JOB_ID_KEY, METADATA_JOB_ID_KEY
from ..models import DATASET_ID_PREFIX, DB_KEY_SEPARATOR, VERSION_ID_PREFIX
from ..parameter_store import ParameterName, get_param
from ..processing_assets_model import ProcessingAssetType, processing_assets_model_with_meta
from ..resources import Resource
from ..s3 import S3_URL_PREFIX
from ..s3_utils import get_bucket_and_key_from_url
from ..step_function_keys import (
    DATASET_ID_KEY,
    DATASET_PREFIX_KEY,
    METADATA_URL_KEY,
    S3_BATCH_RESPONSE_KEY,
    S3_ROLE_ARN_KEY,
    VERSION_ID_KEY,
)
from ..sts import get_account_number
from ..types import JsonObject

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3control import S3ControlClient
    from mypy_boto3_s3control.literals import (
        JobManifestFieldNameType,
        JobManifestFormatType,
        JobReportFormatType,
        JobReportScopeType,
    )
    from mypy_boto3_s3control.type_defs import (
        JobManifestLocationTypeDef,
        JobManifestSpecTypeDef,
        JobManifestTypeDef,
        JobOperationTypeDef,
        JobReportTypeDef,
        LambdaInvokeOperationTypeDef,
    )
else:
    JobManifestLocationTypeDef = (
        JobManifestSpecTypeDef
    ) = (
        JobManifestTypeDef
    ) = (
        JobOperationTypeDef
    ) = JobReportTypeDef = LambdaInvokeOperationTypeDef = dict  # pragma: no mutate
    JobManifestFieldNameType = (
        JobManifestFormatType
    ) = JobReportFormatType = JobReportScopeType = str  # pragma: no mutate
    S3Client = S3ControlClient = object  # pragma: no mutate

LOGGER = get_log()

S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)
S3CONTROL_CLIENT: S3ControlClient = boto3.client("s3control", config=CONFIG)

IMPORT_ASSET_FILE_TASK_ARN = get_param(ParameterName.PROCESSING_IMPORT_ASSET_FILE_FUNCTION_TASK_ARN)
IMPORT_METADATA_FILE_TASK_ARN = get_param(
    ParameterName.PROCESSING_IMPORT_METADATA_FILE_FUNCTION_TASK_ARN
)

STORAGE_BUCKET_ARN = f"arn:aws:s3:::{Resource.STORAGE_BUCKET_NAME.resource_name}"

S3_BATCH_COPY_ROLE_ARN = get_param(ParameterName.PROCESSING_IMPORT_DATASET_ROLE_ARN)

JOB_MANIFEST_FORMAT: JobManifestFormatType = "S3BatchOperations_CSV_20180820"
JOB_MANIFEST_FIELD_NAMES: List[JobManifestFieldNameType] = ["Bucket", "Key"]
JOB_REPORT_FORMAT: JobReportFormatType = "Report_CSV_20180820"
JOB_REPORT_SCOPE: JobReportScopeType = "AllTasks"


class Importer:
    # pylint:disable=too-few-public-methods
    def __init__(  # pylint:disable=too-many-arguments
        self,
        dataset_id: str,
        dataset_prefix: str,
        version_id: str,
        source_bucket_name: str,
        s3_role_arn: str,
    ):
        self.dataset_id = dataset_id
        self.version_id = version_id
        self.source_bucket_name = source_bucket_name
        self.dataset_prefix = dataset_prefix
        self.s3_role_arn = s3_role_arn

    def run(self, task_arn: str, processing_asset_type: ProcessingAssetType) -> str:
        manifest_key = f"manifests/{self.version_id}_{processing_asset_type.value}.csv"
        with smart_open.open(
            f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}/{manifest_key}", "w"
        ) as s3_manifest:
            processing_assets_model = processing_assets_model_with_meta()

            for item in processing_assets_model.query(
                (
                    f"{DATASET_ID_PREFIX}{self.dataset_id}"
                    f"{DB_KEY_SEPARATOR}{VERSION_ID_PREFIX}{self.version_id}"
                ),
                range_key_condition=processing_assets_model.sk.startswith(
                    f"{processing_asset_type.value}{DB_KEY_SEPARATOR}"
                ),
                consistent_read=True,
            ):
                LOGGER.debug(dumps({"Adding file to manifest": item.url}))
                _, key = get_bucket_and_key_from_url(item.url)
                task_parameters = {
                    TARGET_BUCKET_NAME_KEY: Resource.STORAGE_BUCKET_NAME.resource_name,
                    ORIGINAL_KEY_KEY: key,
                    NEW_KEY_KEY: f"{self.dataset_prefix}/{self.version_id}/{basename(key)}",
                    S3_ROLE_ARN_KEY: self.s3_role_arn,
                }
                row = ",".join([self.source_bucket_name, quote(dumps(task_parameters))])
                s3_manifest.write(f"{row}\n")

        manifest_s3_object = S3_CLIENT.head_object(
            Bucket=Resource.STORAGE_BUCKET_NAME.resource_name, Key=manifest_key
        )
        assert "ETag" in manifest_s3_object, manifest_s3_object
        manifest_location_spec = JobManifestLocationTypeDef(
            ObjectArn=f"{STORAGE_BUCKET_ARN}/{manifest_key}", ETag=manifest_s3_object["ETag"]
        )

        account_number = get_account_number()

        # trigger s3 batch copy operation
        response = S3CONTROL_CLIENT.create_job(
            AccountId=account_number,
            ConfirmationRequired=False,
            Operation=JobOperationTypeDef(
                LambdaInvoke=LambdaInvokeOperationTypeDef(FunctionArn=task_arn)
            ),
            Manifest=JobManifestTypeDef(
                Spec=JobManifestSpecTypeDef(
                    Format=JOB_MANIFEST_FORMAT, Fields=JOB_MANIFEST_FIELD_NAMES
                ),
                Location=manifest_location_spec,
            ),
            Report=JobReportTypeDef(
                Enabled=True,
                Bucket=STORAGE_BUCKET_ARN,
                Format=JOB_REPORT_FORMAT,
                Prefix=f"reports/{self.version_id}",
                ReportScope=JOB_REPORT_SCOPE,
            ),
            Priority=1,
            RoleArn=S3_BATCH_COPY_ROLE_ARN,
            ClientRequestToken=uuid4().hex,
        )
        LOGGER.debug(dumps({S3_BATCH_RESPONSE_KEY: response}, default=str))

        return response["JobId"]


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    """Main Lambda entry point."""
    LOGGER.debug(dumps({EVENT_KEY: event}))

    # validate input
    try:
        validate(
            event,
            {
                "type": "object",
                "properties": {
                    DATASET_ID_KEY: {"type": "string"},
                    DATASET_PREFIX_KEY: {"type": "string"},
                    VERSION_ID_KEY: {"type": "string"},
                    METADATA_URL_KEY: {"type": "string"},
                    S3_ROLE_ARN_KEY: {"type": "string"},
                },
                "required": [
                    DATASET_ID_KEY,
                    DATASET_PREFIX_KEY,
                    METADATA_URL_KEY,
                    S3_ROLE_ARN_KEY,
                    VERSION_ID_KEY,
                ],
            },
        )
    except ValidationError as error:
        LOGGER.warning(dumps({ERROR_KEY: error}, default=str))
        return {ERROR_MESSAGE_KEY: error.message}

    source_bucket_name = urlparse(event[METADATA_URL_KEY]).netloc

    importer = Importer(
        event[DATASET_ID_KEY],
        event[DATASET_PREFIX_KEY],
        event[VERSION_ID_KEY],
        source_bucket_name,
        event[S3_ROLE_ARN_KEY],
    )
    asset_job_id = importer.run(IMPORT_ASSET_FILE_TASK_ARN, ProcessingAssetType.DATA)
    metadata_job_id = importer.run(IMPORT_METADATA_FILE_TASK_ARN, ProcessingAssetType.METADATA)

    return {ASSET_JOB_ID_KEY: asset_job_id, METADATA_JOB_ID_KEY: metadata_job_id}



"""
Dataset-versions endpoint Lambda function.
"""
from typing import Callable, Mapping

from ..api_responses import handle_request
from ..types import JsonObject
from .get import get_import_status

REQUEST_HANDLERS: Mapping[str, Callable[[JsonObject], JsonObject]] = {
    "GET": get_import_status,
}


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    return handle_request(event, REQUEST_HANDLERS)



"""Dataset object DynamoDB model."""
from enum import Enum
from os import environ
from typing import Optional, Type

from pynamodb.attributes import UnicodeAttribute
from pynamodb.models import Model

from .aws_keys import AWS_DEFAULT_REGION_KEY
from .parameter_store import ParameterName, get_param


class ProcessingAssetType(Enum):
    DATA = "DATA_ITEM_INDEX"
    METADATA = "METADATA_ITEM_INDEX"


class ProcessingAssetsModelBase(Model):
    pk = UnicodeAttribute(hash_key=True)
    sk = UnicodeAttribute(range_key=True)
    url = UnicodeAttribute()
    multihash = UnicodeAttribute(null=True)


def processing_assets_model_with_meta(
    assets_table_name: Optional[str] = None,
) -> Type[ProcessingAssetsModelBase]:
    if assets_table_name is None:
        assets_table_name = get_param(ParameterName.PROCESSING_ASSETS_TABLE_NAME)

    class ProcessingAssetsModel(ProcessingAssetsModelBase):
        class Meta:  # pylint:disable=too-few-public-methods
            table_name = assets_table_name
            region = environ[AWS_DEFAULT_REGION_KEY]

    return ProcessingAssetsModel



from datetime import datetime
from http import HTTPStatus
from json import dumps, loads
from os import environ
from typing import TYPE_CHECKING

import boto3
from linz_logger import get_log
from slack_sdk.models.blocks import blocks
from slack_sdk.webhook.client import WebhookClient

from ..api_keys import EVENT_KEY
from ..api_responses import success_response
from ..aws_message_attributes import DATA_TYPE_STRING
from ..boto3_config import CONFIG
from ..parameter_store import ParameterName, get_param
from ..step_function import get_import_status_given_arn
from ..step_function_keys import (
    ASSET_UPLOAD_KEY,
    DATASET_PREFIX_KEY,
    INPUT_KEY,
    JOB_STATUS_SUCCEEDED,
    METADATA_UPLOAD_KEY,
    NEW_VERSION_S3_LOCATION,
    OUTPUT_KEY,
    STATUS_KEY,
    UPDATE_DATASET_KEY,
    VALIDATION_KEY,
    VERSION_ID_KEY,
)
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_sns import SNSClient
    from mypy_boto3_sns.type_defs import MessageAttributeValueTypeDef
else:
    # In production we want to avoid depending on a package which has no runtime impact
    SNSClient = object  # pragma: no mutate
    MessageAttributeValueTypeDef = dict  # pragma: no mutate


SLACK_URL_ENV_NAME = "GEOSTORE_SLACK_NOTIFY_URL"

EVENT_DETAIL_KEY = "detail"

MESSAGE_ATTRIBUTE_DATASET_KEY = "dataset_id"
MESSAGE_ATTRIBUTE_STATUS_KEY = "status"

STEP_FUNCTION_ARN_KEY = "executionArn"
STEP_FUNCTION_STARTDATE_KEY = "startDate"
STEP_FUNCTION_STOPDATE_KEY = "stopDate"

WEBHOOK_MESSAGE_BLOCKS_KEY = "blocks"
WEBHOOK_MESSAGE_CHANNEL_KEY = "channel"

SNS_CLIENT: SNSClient = boto3.client("sns", config=CONFIG)
LOGGER = get_log()

BLOCK_MAX_CHAR_LIMIT = 3000  # https://api.slack.com/reference/block-kit/blocks#section


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    LOGGER.debug(dumps({EVENT_KEY: event}))

    if (SLACK_URL_ENV_NAME in environ) and (event[EVENT_DETAIL_KEY][STEP_FUNCTION_STOPDATE_KEY]):
        post_to_slack(event)

    publish_sns_message(event)

    return success_response(HTTPStatus.OK, {})


def publish_sns_message(event: JsonObject) -> None:
    dataset_id = loads(event[EVENT_DETAIL_KEY][INPUT_KEY])[DATASET_PREFIX_KEY]
    SNS_CLIENT.publish(
        TopicArn=get_param(ParameterName.STATUS_SNS_TOPIC_ARN),
        Message=dumps(event),
        MessageAttributes={
            MESSAGE_ATTRIBUTE_DATASET_KEY: MessageAttributeValueTypeDef(
                DataType=DATA_TYPE_STRING, StringValue=dataset_id
            ),
            MESSAGE_ATTRIBUTE_STATUS_KEY: MessageAttributeValueTypeDef(
                DataType=DATA_TYPE_STRING, StringValue=event[EVENT_DETAIL_KEY][STATUS_KEY]
            ),
        },
    )


def post_to_slack(event: JsonObject) -> None:
    event_details = event[EVENT_DETAIL_KEY]
    step_function_input = loads(event_details[INPUT_KEY])
    validation_details = get_import_status_given_arn(event_details[STEP_FUNCTION_ARN_KEY])

    running_time = str(
        datetime.fromtimestamp(event_details[STEP_FUNCTION_STOPDATE_KEY] / 1000)
        - datetime.fromtimestamp(event_details[STEP_FUNCTION_STARTDATE_KEY] / 1000)
    ).split(".", maxsplit=1)[0]

    slack_message_blocks = [
        blocks.HeaderBlock(text="Geostore Dataset Version Import"),
        blocks.DividerBlock(),
        blocks.SectionBlock(text=f"*Status:* {event_details[STATUS_KEY]}"),
        blocks.DividerBlock(),
        blocks.SectionBlock(text=f"*Dataset ID:* `{step_function_input[DATASET_PREFIX_KEY]}`"),
        blocks.SectionBlock(text=f"*Dataset Version ID:* `{step_function_input[VERSION_ID_KEY]}`"),
        blocks.SectionBlock(text=f"*Execution ARN:* `{event_details[STEP_FUNCTION_ARN_KEY]}`"),
        blocks.DividerBlock(),
        blocks.SectionBlock(text=f"*Running Time:* {running_time}"),
        blocks.DividerBlock(),
        blocks.SectionBlock(
            text=format_block("Validation", dumps(validation_details[VALIDATION_KEY]))
        ),
        blocks.SectionBlock(
            text=format_block("Asset Upload", dumps(validation_details[ASSET_UPLOAD_KEY]))
        ),
        blocks.SectionBlock(
            text=format_block("Metadata Upload", dumps(validation_details[METADATA_UPLOAD_KEY]))
        ),
    ]

    if event_details[STATUS_KEY] == JOB_STATUS_SUCCEEDED:
        step_function_output = loads(event_details[OUTPUT_KEY])
        s3_location = step_function_output[UPDATE_DATASET_KEY][NEW_VERSION_S3_LOCATION]
        slack_message_blocks.append(blocks.DividerBlock())
        slack_message_blocks.append(blocks.SectionBlock(text=f"*S3 Location:* `{s3_location}`"))

    response = WebhookClient(environ[SLACK_URL_ENV_NAME]).send(blocks=slack_message_blocks)
    assert response.status_code == HTTPStatus.OK


def format_block(title: str, body: str) -> str:
    """Perform some slack formatting and ensure blocks don't exceed character limit"""
    block_prefix = f"*{title}:* `"
    body_length = BLOCK_MAX_CHAR_LIMIT - len(block_prefix) - 1
    return f"{block_prefix}{body[:body_length]}`"



def decapitalize(key: str) -> str:
    """
    This method will be used to lower case the first character of SQS
    message attributes being received by Lambda to resolve inconsistencies.
    Issue outlined here: https://github.com/boto/boto3/issues/2582
    """
    return f"{key[:1].lower()}{key[1:]}"


MESSAGE_ATTRIBUTE_TYPE_KEY = "type"
MESSAGE_ATTRIBUTE_TYPE_ROOT = "root"
MESSAGE_ATTRIBUTE_TYPE_DATASET = "dataset"
DATA_TYPE_KEY = "DataType"
DATA_TYPE_STRING = "String"
STRING_VALUE_KEY = "StringValue"
STRING_VALUE_KEY_LOWER = decapitalize(STRING_VALUE_KEY)



from typing import TYPE_CHECKING

import boto3
import smart_open

from ..boto3_config import CONFIG
from ..import_dataset_file import get_import_result
from ..s3 import CHUNK_SIZE, S3_URL_PREFIX
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_s3 import S3Client
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3Client = object  # pragma: no mutate

TARGET_S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    return get_import_result(event, importer)


def importer(
    source_bucket_name: str,
    original_key: str,
    target_bucket_name: str,
    new_key: str,
    source_s3_client: S3Client,
) -> None:
    source_response = source_s3_client.get_object(Bucket=source_bucket_name, Key=original_key)

    # TODO: Simplify once boto3 issue #426 is actually fixed pylint:disable=fixme
    with smart_open.open(
        f"{S3_URL_PREFIX}{target_bucket_name}/{new_key}",
        mode="wb",
        transport_params={"client": TARGET_S3_CLIENT},
    ) as target_file:
        for chunk in source_response["Body"].iter_chunks(chunk_size=CHUNK_SIZE):
            target_file.write(chunk)



from os import environ
from typing import Any, Dict, Optional, Tuple, Type

from pynamodb.attributes import UTCDateTimeAttribute, UnicodeAttribute
from pynamodb.indexes import AllProjection, GlobalSecondaryIndex
from pynamodb.models import MetaModel, Model
from ulid import ULID, new

from .aws_keys import AWS_DEFAULT_REGION_KEY
from .clock import now
from .dataset_keys import DATASET_KEY_SEPARATOR
from .models import DATASET_ID_PREFIX, DB_KEY_SEPARATOR
from .parameter_store import ParameterName, get_param


def human_readable_ulid(ulid: ULID) -> str:
    """
    Formats the timestamp part of the ULID as a human readable datetime. Uses "T" as the date/time
    separator as per RFC3339, hyphen as the datetime field separator to ensure broad filesystem
    compatibility, and underscore as the datetime/randomness separator.

    ULIDs have millisecond timestamps, but strftime can only format microseconds, so we need to chop
    off the last three characters.
    """
    datetime_string = ulid.timestamp().datetime.strftime("%Y-%m-%dT%H-%M-%S-%f")[:-3]
    return f"{datetime_string}Z_{ulid.randomness()}"


# TODO: Remove inherit-non-class when astroid is at version 2.6 or later pylint:disable=fixme
class DatasetsTitleIdx(
    GlobalSecondaryIndex["DatasetsModelBase"]
):  # pylint:disable=too-few-public-methods,inherit-non-class
    """Dataset title global index."""

    class Meta:  # pylint:disable=too-few-public-methods
        """Meta class."""

        index_name = "datasets_title"
        read_capacity_units = 1
        write_capacity_units = 1
        projection = AllProjection()

    title = UnicodeAttribute(hash_key=True)


class DatasetsModelBase(Model):
    """Dataset model."""

    id = UnicodeAttribute(
        hash_key=True,
        attr_name="pk",
        default_for_new=lambda: f"{DATASET_ID_PREFIX}{new()}",
    )
    title = UnicodeAttribute()
    created_at = UTCDateTimeAttribute(default_for_new=now)
    updated_at = UTCDateTimeAttribute(default=now)

    datasets_title_idx: DatasetsTitleIdx

    def as_dict(self) -> Dict[str, Any]:
        serialized = self.serialize()
        result: Dict[str, Any] = {key: value["S"] for key, value in serialized.items()}
        result["id"] = self.dataset_id
        return result

    @property
    def dataset_id(self) -> str:
        """Dataset ID value."""
        return str(self.id).split(DB_KEY_SEPARATOR)[1]

    @property
    def dataset_prefix(self) -> str:
        """Dataset prefix value."""
        return f"{self.title}{DATASET_KEY_SEPARATOR}{self.dataset_id}"


class DatasetsModelMeta(MetaModel):
    def __new__(
        cls,
        name: str,
        bases: Tuple[Type[object], ...],
        namespace: Dict[str, Any],
        discriminator: Optional[Any] = None,
    ) -> "DatasetsModelMeta":
        namespace["Meta"] = type(
            "Meta",
            (),
            {
                "table_name": get_param(ParameterName.STORAGE_DATASETS_TABLE_NAME),
                "region": environ[AWS_DEFAULT_REGION_KEY],
            },
        )
        klass: "DatasetsModelMeta" = MetaModel.__new__(  # type: ignore[no-untyped-call]
            cls, name, bases, namespace, discriminator=discriminator
        )
        return klass


def datasets_model_with_meta() -> Type[DatasetsModelBase]:
    class DatasetModel(DatasetsModelBase, metaclass=DatasetsModelMeta):
        datasets_title_idx = DatasetsTitleIdx()

    return DatasetModel



from json import dumps
from typing import TYPE_CHECKING

import boto3
from linz_logger import get_log
from pystac import read_file
from pystac.catalog import Catalog, CatalogType
from pystac.collection import Collection
from pystac.item import Item
from pystac.layout import HrefLayoutStrategy
from pystac.stac_io import StacIO

from ..api_keys import EVENT_KEY
from ..aws_keys import BODY_KEY
from ..aws_message_attributes import (
    MESSAGE_ATTRIBUTE_TYPE_DATASET,
    MESSAGE_ATTRIBUTE_TYPE_KEY,
    MESSAGE_ATTRIBUTE_TYPE_ROOT,
    STRING_VALUE_KEY_LOWER,
)
from ..boto3_config import CONFIG
from ..pystac_io_methods import S3StacIO
from ..resources import Resource
from ..s3 import S3_URL_PREFIX
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_s3 import S3Client
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3Client = object  # pragma: no mutate

S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)

ROOT_CATALOG_ID = "root_catalog"
ROOT_CATALOG_TITLE = "LINZ Geostore"
ROOT_CATALOG_DESCRIPTION = (
    "The LINZ Geospatial Data Store (Geostore) contains all the important "
    "geospatial data held by Land Information New Zealand (LINZ).<br/>"
    "Please browse this catalog to find and access our data."
)
CATALOG_FILENAME = "catalog.json"
CONTENTS_KEY = "Contents"
RECORDS_KEY = "Records"
MESSAGE_ATTRIBUTES_KEY = "messageAttributes"

LOGGER = get_log()

StacIO.set_default(S3StacIO)


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    """Main Lambda entry point."""

    LOGGER.debug(dumps({EVENT_KEY: event}))

    for message in event[RECORDS_KEY]:
        if (
            message[MESSAGE_ATTRIBUTES_KEY][MESSAGE_ATTRIBUTE_TYPE_KEY][STRING_VALUE_KEY_LOWER]
            == MESSAGE_ATTRIBUTE_TYPE_ROOT
        ):
            handle_root(message[BODY_KEY])
        elif (
            message[MESSAGE_ATTRIBUTES_KEY][MESSAGE_ATTRIBUTE_TYPE_KEY][STRING_VALUE_KEY_LOWER]
            == MESSAGE_ATTRIBUTE_TYPE_DATASET
        ):
            handle_dataset(message[BODY_KEY])
        else:
            raise UnhandledSQSMessageException("Unhandled SQS message type")

    return {}


class UnhandledSQSMessageException(Exception):
    pass


class GeostoreSTACLayoutStrategy(HrefLayoutStrategy):
    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        return str(cat.get_self_href())

    def get_collection_href(self, col: Collection, parent_dir: str, is_root: bool) -> str:
        assert not is_root
        return str(col.get_self_href())

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        return str(item.get_self_href())


def handle_dataset(version_metadata_key: str) -> None:
    """Handle writing a new dataset version to the dataset catalog"""
    storage_bucket_path = f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}"
    dataset_prefix = version_metadata_key.split("/", maxsplit=1)[0]
    dataset_catalog = Catalog.from_file(
        f"{storage_bucket_path}/{dataset_prefix}/{CATALOG_FILENAME}"
    )

    dataset_version_metadata = read_file(f"{storage_bucket_path}/{version_metadata_key}")
    assert isinstance(dataset_version_metadata, (Catalog, Collection))
    dataset_catalog.add_child(child=dataset_version_metadata, strategy=GeostoreSTACLayoutStrategy())

    dataset_catalog.normalize_hrefs(
        f"{storage_bucket_path}/{dataset_prefix}", strategy=GeostoreSTACLayoutStrategy()
    )
    dataset_catalog.save(catalog_type=CatalogType.SELF_CONTAINED)


def handle_root(dataset_prefix: str) -> None:
    """Handle writing a new dataset to the root catalog"""
    results = S3_CLIENT.list_objects(
        Bucket=Resource.STORAGE_BUCKET_NAME.resource_name, Prefix=CATALOG_FILENAME
    )

    # create root catalog if it doesn't exist
    if CONTENTS_KEY in results:
        root_catalog = Catalog.from_file(
            f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}/{CATALOG_FILENAME}"
        )

    else:
        root_catalog = Catalog(
            id=ROOT_CATALOG_ID,
            title=ROOT_CATALOG_TITLE,
            description=ROOT_CATALOG_DESCRIPTION,
            catalog_type=CatalogType.SELF_CONTAINED,
        )
        root_catalog.set_self_href(
            f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}/{CATALOG_FILENAME}"
        )

    dataset_path = f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}/{dataset_prefix}"
    dataset_catalog = Catalog.from_file(f"{dataset_path}/{CATALOG_FILENAME}")

    root_catalog.add_child(dataset_catalog, strategy=GeostoreSTACLayoutStrategy())
    root_catalog.normalize_hrefs(
        f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}",
        strategy=GeostoreSTACLayoutStrategy(),
    )

    root_catalog.save(catalog_type=CatalogType.SELF_CONTAINED)



"""Dataset versions handler function."""
from datetime import datetime
from http import HTTPStatus
from json import dumps
from typing import TYPE_CHECKING

import boto3
from jsonschema import ValidationError, validate
from linz_logger import get_log
from pynamodb.exceptions import DoesNotExist
from ulid import from_timestamp

from ..api_keys import EVENT_KEY
from ..api_responses import error_response, success_response
from ..boto3_config import CONFIG
from ..datasets_model import datasets_model_with_meta, human_readable_ulid
from ..error_response_keys import ERROR_KEY
from ..models import DATASET_ID_PREFIX
from ..parameter_store import ParameterName, get_param
from ..step_function_keys import (
    DATASET_ID_KEY,
    DATASET_ID_SHORT_KEY,
    DATASET_PREFIX_KEY,
    EXECUTION_ARN_KEY,
    METADATA_URL_KEY,
    NOW_KEY,
    S3_ROLE_ARN_KEY,
    VERSION_ID_KEY,
)
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_stepfunctions import SFNClient
else:
    # In production we want to avoid depending on a package which has no runtime impact
    SFNClient = object  # pragma: no mutate

STEP_FUNCTIONS_CLIENT: SFNClient = boto3.client("stepfunctions", config=CONFIG)
LOGGER = get_log()


def create_dataset_version(body: JsonObject) -> JsonObject:

    LOGGER.debug(dumps({EVENT_KEY: body}))

    body_schema = {
        "type": "object",
        "properties": {
            DATASET_ID_SHORT_KEY: {"type": "string"},
            METADATA_URL_KEY: {"type": "string"},
            NOW_KEY: {"type": "string", "format": "date-time"},
            S3_ROLE_ARN_KEY: {"type": "string"},
        },
        "required": [DATASET_ID_SHORT_KEY, METADATA_URL_KEY, S3_ROLE_ARN_KEY],
    }

    # validate input
    try:
        validate(body, body_schema)
    except ValidationError as err:
        LOGGER.warning(dumps({ERROR_KEY: err}, default=str))
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    datasets_model_class = datasets_model_with_meta()

    # validate dataset exists
    try:
        dataset = datasets_model_class.get(
            hash_key=f"{DATASET_ID_PREFIX}{body[DATASET_ID_SHORT_KEY]}", consistent_read=True
        )
    except DoesNotExist as err:
        LOGGER.warning(dumps({ERROR_KEY: err}, default=str))
        return error_response(
            HTTPStatus.NOT_FOUND, f"dataset '{body[DATASET_ID_SHORT_KEY]}' could not be found"
        )

    now = datetime.fromisoformat(body.get(NOW_KEY, datetime.utcnow().isoformat()))
    dataset_version_id = human_readable_ulid(from_timestamp(now))

    # execute step function
    step_functions_input = {
        DATASET_ID_KEY: dataset.dataset_id,
        DATASET_PREFIX_KEY: dataset.dataset_prefix,
        VERSION_ID_KEY: dataset_version_id,
        METADATA_URL_KEY: body[METADATA_URL_KEY],
        S3_ROLE_ARN_KEY: body[S3_ROLE_ARN_KEY],
    }
    state_machine_arn = get_param(
        ParameterName.PROCESSING_DATASET_VERSION_CREATION_STEP_FUNCTION_ARN
    )

    step_functions_response = STEP_FUNCTIONS_CLIENT.start_execution(
        stateMachineArn=state_machine_arn,
        name=dataset_version_id,
        input=dumps(step_functions_input),
    )

    LOGGER.debug(dumps({"response": step_functions_response}, default=str))

    # return arn of executing process
    return success_response(
        HTTPStatus.CREATED,
        {
            VERSION_ID_KEY: dataset_version_id,
            EXECUTION_ARN_KEY: step_functions_response["executionArn"],
        },
    )



"""Create dataset function."""
from http import HTTPStatus
from string import ascii_letters, digits
from typing import TYPE_CHECKING

import boto3
from jsonschema import ValidationError, validate
from pystac.catalog import Catalog, CatalogType
from pystac.stac_io import StacIO

from ..api_responses import error_response, success_response
from ..aws_message_attributes import (
    DATA_TYPE_STRING,
    MESSAGE_ATTRIBUTE_TYPE_KEY,
    MESSAGE_ATTRIBUTE_TYPE_ROOT,
)
from ..datasets_model import datasets_model_with_meta
from ..parameter_store import ParameterName, get_param
from ..pystac_io_methods import S3StacIO
from ..resources import Resource
from ..s3 import S3_URL_PREFIX
from ..stac_format import STAC_DESCRIPTION_KEY, STAC_ID_KEY, STAC_TITLE_KEY
from ..step_function_keys import DESCRIPTION_KEY, TITLE_KEY
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_sqs import SQSServiceResource
    from mypy_boto3_sqs.type_defs import MessageAttributeValueTypeDef
else:
    # In production we want to avoid depending on a package which has no runtime impact
    SQSServiceResource = object  # pragma: no mutate
    MessageAttributeValueTypeDef = dict  # pragma: no mutate

TITLE_CHARACTERS = f"{ascii_letters}{digits}_-"
TITLE_PATTERN = f"^[{TITLE_CHARACTERS}]+$"

SQS_RESOURCE: SQSServiceResource = boto3.resource("sqs")

StacIO.set_default(S3StacIO)


def create_dataset(body: JsonObject) -> JsonObject:
    """POST: Create Dataset."""

    body_schema = {
        "type": "object",
        "properties": {
            TITLE_KEY: {"type": "string", "pattern": TITLE_PATTERN},
            DESCRIPTION_KEY: {"type": "string"},
        },
        "required": [TITLE_KEY, DESCRIPTION_KEY],
    }

    # request body validation
    try:
        validate(body, body_schema)
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    # check for duplicate type/title
    datasets_model_class = datasets_model_with_meta()
    dataset_title = body[TITLE_KEY]
    if datasets_model_class.datasets_title_idx.count(hash_key=dataset_title):
        return error_response(HTTPStatus.CONFLICT, f"dataset '{dataset_title}' already exists")

    # create dataset
    dataset = datasets_model_class(title=dataset_title)
    dataset.save()
    dataset.refresh(consistent_read=True)

    # create dataset catalog
    dataset_catalog = Catalog(
        **{
            STAC_ID_KEY: dataset.dataset_prefix,
            STAC_DESCRIPTION_KEY: body[DESCRIPTION_KEY],
            STAC_TITLE_KEY: dataset_title,
        },
        catalog_type=CatalogType.SELF_CONTAINED,
    )
    dataset_catalog.normalize_hrefs(
        f"{S3_URL_PREFIX}{Resource.STORAGE_BUCKET_NAME.resource_name}/{dataset.dataset_prefix}"
    )
    dataset_catalog.save()

    # add reference to root catalog
    SQS_RESOURCE.get_queue_by_name(
        QueueName=get_param(ParameterName.UPDATE_CATALOG_MESSAGE_QUEUE_NAME)
    ).send_message(
        MessageBody=dataset.dataset_prefix,
        MessageAttributes={
            MESSAGE_ATTRIBUTE_TYPE_KEY: MessageAttributeValueTypeDef(
                DataType=DATA_TYPE_STRING, StringValue=MESSAGE_ATTRIBUTE_TYPE_ROOT
            )
        },
    )

    # return response
    resp_body = dataset.as_dict()

    return success_response(HTTPStatus.CREATED, resp_body)



from json import dumps, loads
from typing import TYPE_CHECKING, Callable, Optional
from urllib.parse import unquote_plus

from botocore.exceptions import ClientError
from linz_logger import get_log

from .aws_response import AWS_CODE_REQUEST_TIMEOUT
from .import_dataset_keys import NEW_KEY_KEY, ORIGINAL_KEY_KEY, TARGET_BUCKET_NAME_KEY
from .s3 import get_s3_client_for_role
from .step_function_keys import S3_ROLE_ARN_KEY
from .types import JsonObject

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3.type_defs import PutObjectOutputTypeDef
else:
    PutObjectOutputTypeDef = JsonObject  # pragma: no mutate
    S3Client = object  # pragma: no mutate

INVOCATION_ID_KEY = "invocationId"
INVOCATION_SCHEMA_VERSION_KEY = "invocationSchemaVersion"
RESULTS_KEY = "results"
RESULT_CODE_KEY = "resultCode"
RESULT_STRING_KEY = "resultString"
S3_BUCKET_ARN_KEY = "s3BucketArn"
S3_KEY_KEY = "s3Key"
TASKS_KEY = "tasks"
TASK_ID_KEY = "taskId"
TREAT_MISSING_KEYS_AS_KEY = "treatMissingKeysAs"

RESULT_CODE_PERMANENT_FAILURE = "PermanentFailure"
RESULT_CODE_SUCCEEDED = "Succeeded"
RESULT_CODE_TEMPORARY_FAILURE = "TemporaryFailure"

EXCEPTION_PREFIX = "Exception"
RETRY_RESULT_STRING = "Retry request to Amazon S3 due to timeout."

LOGGER = get_log()


def get_import_result(
    event: JsonObject,
    importer: Callable[[str, str, str, str, S3Client], Optional[PutObjectOutputTypeDef]],
) -> JsonObject:
    LOGGER.debug(dumps(event))

    task = event[TASKS_KEY][0]
    source_bucket_name = task[S3_BUCKET_ARN_KEY].split(":::", maxsplit=1)[-1]
    parameters = loads(unquote_plus(task[S3_KEY_KEY]))
    source_s3_client = get_s3_client_for_role(parameters[S3_ROLE_ARN_KEY])

    try:
        response = importer(
            source_bucket_name,
            parameters[ORIGINAL_KEY_KEY],
            parameters[TARGET_BUCKET_NAME_KEY],
            parameters[NEW_KEY_KEY],
            source_s3_client,
        )
        result_code = RESULT_CODE_SUCCEEDED
        result_string = str(response)
    except ClientError as error:
        error_code = error.response["Error"]["Code"]
        if error_code == AWS_CODE_REQUEST_TIMEOUT:
            result_code = RESULT_CODE_TEMPORARY_FAILURE
            result_string = RETRY_RESULT_STRING
        else:
            result_code = RESULT_CODE_PERMANENT_FAILURE
            error_message = error.response["Error"]["Message"]
            result_string = f"{error_code} when calling {error.operation_name}: {error_message}"
    except Exception as error:  # pylint:disable=broad-except
        result_code = RESULT_CODE_PERMANENT_FAILURE
        result_string = f"{EXCEPTION_PREFIX}: {error}"

    result = {
        INVOCATION_SCHEMA_VERSION_KEY: event[INVOCATION_SCHEMA_VERSION_KEY],
        TREAT_MISSING_KEYS_AS_KEY: RESULT_CODE_PERMANENT_FAILURE,
        INVOCATION_ID_KEY: event[INVOCATION_ID_KEY],
        RESULTS_KEY: [
            {
                TASK_ID_KEY: task[TASK_ID_KEY],
                RESULT_CODE_KEY: result_code,
                RESULT_STRING_KEY: result_string,
            }
        ],
    }
    LOGGER.debug(dumps(result))
    return result



from json import dumps
from logging import Logger
from os import environ
from urllib.parse import urlparse

from botocore.exceptions import ClientError
from multihash import FUNCS, decode

from ..api_keys import MESSAGE_KEY, SUCCESS_KEY
from ..check import Check
from ..error_response_keys import ERROR_KEY
from ..processing_assets_model import processing_assets_model_with_meta
from ..s3 import CHUNK_SIZE, get_s3_client_for_role
from ..types import JsonObject
from ..validation_results_model import ValidationResult, ValidationResultFactory

ARRAY_INDEX_VARIABLE_NAME = "AWS_BATCH_JOB_ARRAY_INDEX"


class ChecksumMismatchError(Exception):
    def __init__(self, actual_hex_digest: str):
        super().__init__()

        self.actual_hex_digest = actual_hex_digest


class ChecksumValidator:
    def __init__(
        self,
        processing_assets_table_name: str,
        validation_result_factory: ValidationResultFactory,
        s3_role_arn: str,
        logger: Logger,
    ):
        self.validation_result_factory = validation_result_factory
        self.logger = logger

        self.processing_assets_model = processing_assets_model_with_meta(
            processing_assets_table_name
        )

        self.s3_client = get_s3_client_for_role(s3_role_arn)

    def log_failure(self, content: JsonObject) -> None:
        self.logger.error(dumps({SUCCESS_KEY: False, **content}))

    def validate(self, hash_key: str, range_key: str) -> None:

        try:
            item = self.processing_assets_model.get(hash_key, range_key=range_key)
        except self.processing_assets_model.DoesNotExist:
            self.log_failure(
                {
                    ERROR_KEY: {MESSAGE_KEY: "Item does not exist"},
                    "parameters": {"hash_key": hash_key, "range_key": range_key},
                }
            )
            raise

        try:
            self.validate_url_multihash(item.url, item.multihash)
        except ChecksumMismatchError as error:
            content = {
                MESSAGE_KEY: f"Checksum mismatch: expected {item.multihash[4:]},"
                f" got {error.actual_hex_digest}"
            }
            self.log_failure(content)
            self.validation_result_factory.save(
                item.url, Check.CHECKSUM, ValidationResult.FAILED, details=content
            )
        else:
            self.logger.info(dumps({SUCCESS_KEY: True, MESSAGE_KEY: ""}))
            self.validation_result_factory.save(item.url, Check.CHECKSUM, ValidationResult.PASSED)

    def validate_url_multihash(self, url: str, hex_multihash: str) -> None:
        parsed_url = urlparse(url)
        bucket = parsed_url.netloc
        key = parsed_url.path.lstrip("/")
        try:
            url_stream = self.s3_client.get_object(Bucket=bucket, Key=key)["Body"]
        except ClientError as error:
            self.validation_result_factory.save(
                url,
                Check.STAGING_ACCESS,
                ValidationResult.FAILED,
                details={MESSAGE_KEY: str(error)},
            )
            raise

        checksum_function_code = int(hex_multihash[:2], 16)
        checksum_function = FUNCS[checksum_function_code]

        file_digest = checksum_function()
        for chunk in url_stream.iter_chunks(chunk_size=CHUNK_SIZE):
            file_digest.update(chunk)

        if file_digest.digest() != decode(bytes.fromhex(hex_multihash)):
            raise ChecksumMismatchError(file_digest.hexdigest())


def get_job_offset() -> int:
    return int(environ.get(ARRAY_INDEX_VARIABLE_NAME, 0))



from jsonschema import validate

from ..models import DATASET_ID_PREFIX, DB_KEY_SEPARATOR, VERSION_ID_PREFIX
from ..parameter_store import ParameterName, get_param
from ..processing_assets_model import ProcessingAssetType, processing_assets_model_with_meta
from ..step_function_keys import DATASET_ID_KEY, METADATA_URL_KEY, VERSION_ID_KEY
from ..types import JsonObject

MAX_ITERATION_SIZE = 10_000

ASSETS_TABLE_NAME_KEY = "assets_table_name"
CONTENT_KEY = "content"
FIRST_ITEM_KEY = "first_item"
ITERATION_SIZE_KEY = "iteration_size"
NEXT_ITEM_KEY = "next_item"
RESULTS_TABLE_NAME_KEY = "results_table_name"

EVENT_SCHEMA = {
    "type": "object",
    "properties": {
        CONTENT_KEY: {
            "type": "object",
            "properties": {
                FIRST_ITEM_KEY: {"type": "string", "pattern": r"^\d+$"},
                ITERATION_SIZE_KEY: {
                    "type": "integer",
                    "minimum": 1,
                    "maximum": MAX_ITERATION_SIZE,
                },
                NEXT_ITEM_KEY: {
                    "type": "integer",
                    "minimum": MAX_ITERATION_SIZE,
                    "multipleOf": MAX_ITERATION_SIZE,
                },
            },
            "required": [FIRST_ITEM_KEY, ITERATION_SIZE_KEY, NEXT_ITEM_KEY],
            "additionalProperties": False,
        },
        DATASET_ID_KEY: {"type": "string"},
        METADATA_URL_KEY: {"type": "string"},
        VERSION_ID_KEY: {"type": "string"},
    },
    "required": [DATASET_ID_KEY, METADATA_URL_KEY, VERSION_ID_KEY],
    "additionalProperties": True,
}


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    validate(event, EVENT_SCHEMA)

    if CONTENT_KEY in event.keys():
        assert int(event[CONTENT_KEY][FIRST_ITEM_KEY]) % MAX_ITERATION_SIZE == 0
        first_item_index = event[CONTENT_KEY][NEXT_ITEM_KEY]
    else:
        first_item_index = 0

    dataset_id = event[DATASET_ID_KEY]
    version_id = event[VERSION_ID_KEY]

    processing_assets_model = processing_assets_model_with_meta()

    asset_count = processing_assets_model.count(
        hash_key=(
            f"{DATASET_ID_PREFIX}{dataset_id}{DB_KEY_SEPARATOR}{VERSION_ID_PREFIX}{version_id}"
        ),
        range_key_condition=processing_assets_model.sk.startswith(
            f"{ProcessingAssetType.DATA.value}{DB_KEY_SEPARATOR}"
        ),
    )

    remaining_assets = asset_count - first_item_index
    if remaining_assets > MAX_ITERATION_SIZE:
        next_item_index = first_item_index + MAX_ITERATION_SIZE
        iteration_size = MAX_ITERATION_SIZE
    else:
        next_item_index = -1
        iteration_size = remaining_assets

    return {
        FIRST_ITEM_KEY: str(first_item_index),
        ITERATION_SIZE_KEY: iteration_size,
        NEXT_ITEM_KEY: next_item_index,
        ASSETS_TABLE_NAME_KEY: get_param(ParameterName.PROCESSING_ASSETS_TABLE_NAME),
        RESULTS_TABLE_NAME_KEY: get_param(ParameterName.STORAGE_VALIDATION_RESULTS_TABLE_NAME),
    }



"""Update dataset function."""
from http import HTTPStatus

from jsonschema import ValidationError, validate
from pynamodb.exceptions import DoesNotExist

from ..api_responses import error_response, success_response
from ..datasets_model import DatasetsModelBase, datasets_model_with_meta
from ..models import DATASET_ID_PREFIX
from ..step_function_keys import DATASET_ID_SHORT_KEY, TITLE_KEY
from ..types import JsonObject


def update_dataset(body: JsonObject) -> JsonObject:
    """PATCH: Update Dataset."""

    body_schema = {
        "type": "object",
        "properties": {DATASET_ID_SHORT_KEY: {"type": "string"}, TITLE_KEY: {"type": "string"}},
        "required": [DATASET_ID_SHORT_KEY, TITLE_KEY],
    }

    # request body validation
    try:
        validate(body, body_schema)
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    # check for duplicate type/title
    datasets_model_class = datasets_model_with_meta()
    dataset_title = body[TITLE_KEY]
    if datasets_model_class.datasets_title_idx.count(hash_key=dataset_title):
        return error_response(HTTPStatus.CONFLICT, f"dataset '{dataset_title}' already exists")

    # get dataset to update
    dataset_id = body[DATASET_ID_SHORT_KEY]
    try:
        dataset = datasets_model_class.get(
            hash_key=f"{DATASET_ID_PREFIX}{dataset_id}", consistent_read=True
        )
    except DoesNotExist:
        return error_response(HTTPStatus.NOT_FOUND, f"dataset '{dataset_id}' does not exist")

    # update dataset
    update_dataset_attributes(dataset, body)
    dataset.save()
    dataset.refresh(consistent_read=True)

    # return response
    resp_body = dataset.as_dict()

    return success_response(HTTPStatus.OK, resp_body)


def update_dataset_attributes(dataset: DatasetsModelBase, req_body: JsonObject) -> None:
    for attr in DatasetsModelBase.get_attributes():
        if attr in req_body and attr != "id":
            setattr(dataset, attr, req_body[attr])



from json import dumps
from os.path import basename
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import boto3
from jsonschema import ValidationError, validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY
from ..aws_message_attributes import (
    DATA_TYPE_STRING,
    MESSAGE_ATTRIBUTE_TYPE_DATASET,
    MESSAGE_ATTRIBUTE_TYPE_KEY,
)
from ..error_response_keys import ERROR_KEY, ERROR_MESSAGE_KEY
from ..parameter_store import ParameterName, get_param
from ..resources import Resource
from ..s3 import S3_URL_PREFIX
from ..step_function_keys import (
    DATASET_ID_KEY,
    DATASET_PREFIX_KEY,
    METADATA_URL_KEY,
    NEW_VERSION_S3_LOCATION,
    VERSION_ID_KEY,
)
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_sqs import SQSServiceResource
    from mypy_boto3_sqs.type_defs import MessageAttributeValueTypeDef
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3Client = SQSServiceResource = object  # pragma: no mutate
    MessageAttributeValueTypeDef = dict  # pragma: no mutate

LOGGER = get_log()
SQS_RESOURCE: SQSServiceResource = boto3.resource("sqs")


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    """Main Lambda entry point."""
    LOGGER.debug(dumps({EVENT_KEY: event}))

    # validate input
    try:
        validate(
            event,
            {
                "type": "object",
                "properties": {
                    DATASET_ID_KEY: {"type": "string"},
                    DATASET_PREFIX_KEY: {"type": "string"},
                    VERSION_ID_KEY: {"type": "string"},
                    METADATA_URL_KEY: {"type": "string"},
                },
                "required": [DATASET_ID_KEY, DATASET_PREFIX_KEY, METADATA_URL_KEY, VERSION_ID_KEY],
            },
        )
    except ValidationError as error:
        LOGGER.warning(dumps({ERROR_KEY: error}, default=str))
        return {ERROR_MESSAGE_KEY: error.message}

    new_version_metadata_key = (
        f"{event[DATASET_PREFIX_KEY]}/{event[VERSION_ID_KEY]}/"
        f"{basename(urlparse(event[METADATA_URL_KEY]).path[1:])}"
    )

    # add reference to root catalog
    SQS_RESOURCE.get_queue_by_name(
        QueueName=get_param(ParameterName.UPDATE_CATALOG_MESSAGE_QUEUE_NAME)
    ).send_message(
        MessageBody=new_version_metadata_key,
        MessageAttributes={
            MESSAGE_ATTRIBUTE_TYPE_KEY: MessageAttributeValueTypeDef(
                DataType=DATA_TYPE_STRING, StringValue=MESSAGE_ATTRIBUTE_TYPE_DATASET
            )
        },
    )

    return {
        NEW_VERSION_S3_LOCATION: f"{S3_URL_PREFIX}"
        f"{Resource.STORAGE_BUCKET_NAME.resource_name}/"
        f"{new_version_metadata_key}"
    }



from typing import TYPE_CHECKING
from uuid import uuid4

import boto3

from .boto3_config import CONFIG
from .environment import environment_name

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_sts import STSClient
else:
    S3Client = STSClient = object  # pragma: no mutate


S3_SCHEMA = "s3"
S3_URL_PREFIX = f"{S3_SCHEMA}://"

CHUNK_SIZE = 1024

STS_CLIENT: STSClient = boto3.client("sts", config=CONFIG)


def get_s3_client_for_role(role_arn: str) -> S3Client:
    assume_role_response = STS_CLIENT.assume_role(
        RoleArn=role_arn, RoleSessionName=f"{environment_name()}_Geostore_{uuid4()}"
    )
    credentials = assume_role_response["Credentials"]
    client: S3Client = boto3.client(
        "s3",
        config=CONFIG,
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )
    return client



"""
Dataset endpoint Lambda function.
"""
from typing import Callable, MutableMapping

from ..api_responses import handle_request
from ..types import JsonObject
from .create import create_dataset
from .delete import delete_dataset
from .get import handle_get
from .update import update_dataset

REQUEST_HANDLERS: MutableMapping[str, Callable[[JsonObject], JsonObject]] = {
    "DELETE": delete_dataset,
    "GET": handle_get,
    "PATCH": update_dataset,
    "POST": create_dataset,
}


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    return handle_request(event, REQUEST_HANDLERS)



from functools import lru_cache
from typing import TYPE_CHECKING

import boto3

from .boto3_config import CONFIG

if TYPE_CHECKING:
    from mypy_boto3_sts import STSClient
else:
    STSClient = object  # pragma: no mutate

STS_CLIENT: STSClient = boto3.client("sts", config=CONFIG)


@lru_cache
def get_account_number() -> str:
    caller_identity = STS_CLIENT.get_caller_identity()
    assert "Account" in caller_identity, caller_identity
    return caller_identity["Account"]



from enum import Enum, auto
from functools import lru_cache
from json import dumps
from typing import TYPE_CHECKING, Sequence

import boto3
from linz_logger import get_log

from .boto3_config import CONFIG
from .environment import environment_name
from .error_response_keys import ERROR_KEY

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_ssm import SSMClient
else:
    # In production we want to avoid depending on a package which has no runtime impact
    SSMClient = object  # pragma: no mutate

LOGGER = get_log()
SSM_CLIENT: SSMClient = boto3.client("ssm", config=CONFIG)


class ParameterName(Enum):
    # Use @staticmethod instead of all the ignores on the next line once we move to Python 3.9
    # <https://github.com/python/mypy/issues/7591>.
    def _generate_next_value_(  # type: ignore[misc,override] # pylint:disable=no-self-argument,no-member
        name: str, _start: int, _count: int, _last_values: Sequence[str]
    ) -> str:
        return f"/{environment_name()}/{name.lower()}"

    PROCESSING_ASSETS_TABLE_NAME = auto()
    PROCESSING_DATASET_VERSION_CREATION_STEP_FUNCTION_ARN = auto()
    PROCESSING_IMPORT_ASSET_FILE_FUNCTION_TASK_ARN = auto()
    PROCESSING_IMPORT_DATASET_ROLE_ARN = auto()
    PROCESSING_IMPORT_METADATA_FILE_FUNCTION_TASK_ARN = auto()
    UPDATE_CATALOG_MESSAGE_QUEUE_NAME = auto()
    STATUS_SNS_TOPIC_ARN = auto()
    STORAGE_DATASETS_TABLE_NAME = auto()
    STORAGE_VALIDATION_RESULTS_TABLE_NAME = auto()


@lru_cache
def get_param(parameter: ParameterName) -> str:
    try:
        return SSM_CLIENT.get_parameter(Name=parameter.value)["Parameter"]["Value"]
    except SSM_CLIENT.exceptions.ParameterNotFound:
        LOGGER.error(dumps({ERROR_KEY: f"Parameter not found: “{parameter.value}”"}))
        raise



from functools import lru_cache
from json import JSONDecodeError, dumps, load
from os.path import dirname
from typing import Any, Callable, Dict, List, Tuple

from botocore.exceptions import ClientError
from botocore.response import StreamingBody
from jsonschema import Draft7Validator, ValidationError
from linz_logger import get_log

from ..api_keys import MESSAGE_KEY, SUCCESS_KEY
from ..check import Check
from ..models import DB_KEY_SEPARATOR
from ..processing_assets_model import ProcessingAssetType, processing_assets_model_with_meta
from ..s3 import S3_URL_PREFIX
from ..stac_format import (
    STAC_ASSETS_KEY,
    STAC_FILE_CHECKSUM_KEY,
    STAC_HREF_KEY,
    STAC_LINKS_KEY,
    STAC_TYPE_CATALOG,
    STAC_TYPE_COLLECTION,
    STAC_TYPE_ITEM,
    STAC_TYPE_KEY,
)
from ..types import JsonObject
from ..validation_results_model import ValidationResult, ValidationResultFactory
from .stac_validators import (
    STACCatalogSchemaValidator,
    STACCollectionSchemaValidator,
    STACItemSchemaValidator,
)

NO_ASSETS_FOUND_ERROR_MESSAGE = "No assets found in dataset"

LOGGER = get_log()

STAC_TYPE_VALIDATION_MAP: Dict[str, Draft7Validator] = {
    STAC_TYPE_CATALOG: STACCatalogSchemaValidator,
    STAC_TYPE_COLLECTION: STACCollectionSchemaValidator,
    STAC_TYPE_ITEM: STACItemSchemaValidator,
}

PROCESSING_ASSET_ASSET_KEY = "asset"
PROCESSING_ASSET_MULTIHASH_KEY = "multihash"
PROCESSING_ASSET_URL_KEY = "url"

EXPLICITLY_RELATIVE_PATH_PREFIX = "./"


@lru_cache
def maybe_convert_relative_url_to_absolute(url_or_path: str, parent_url: str) -> str:
    if url_or_path.startswith(S3_URL_PREFIX):
        return url_or_path

    if url_or_path.startswith(EXPLICITLY_RELATIVE_PATH_PREFIX):
        url_or_path = url_or_path[len(EXPLICITLY_RELATIVE_PATH_PREFIX) :]

    return f"{dirname(parent_url)}/{url_or_path}"


class STACDatasetValidator:
    def __init__(
        self,
        hash_key: str,
        url_reader: Callable[[str], StreamingBody],
        validation_result_factory: ValidationResultFactory,
    ):
        self.hash_key = hash_key
        self.url_reader = url_reader
        self.validation_result_factory = validation_result_factory

        self.traversed_urls: List[str] = []
        self.dataset_assets: List[Dict[str, str]] = []
        self.dataset_metadata: List[Dict[str, str]] = []

        self.processing_assets_model = processing_assets_model_with_meta()

    def run(self, metadata_url: str) -> None:
        if metadata_url[:5] != S3_URL_PREFIX:
            error_message = f"URL doesn't start with “{S3_URL_PREFIX}”: “{metadata_url}”"
            self.validation_result_factory.save(
                metadata_url,
                Check.NON_S3_URL,
                ValidationResult.FAILED,
                details={MESSAGE_KEY: error_message},
            )
            LOGGER.error(dumps({SUCCESS_KEY: False, MESSAGE_KEY: error_message}))
            return

        try:
            self.validate(metadata_url)
        except (ValidationError, ClientError, JSONDecodeError) as error:
            LOGGER.error(dumps({SUCCESS_KEY: False, MESSAGE_KEY: str(error)}))
            return

        if not self.dataset_assets:
            error_details = {MESSAGE_KEY: NO_ASSETS_FOUND_ERROR_MESSAGE}
            self.validation_result_factory.save(
                metadata_url,
                Check.ASSETS_IN_DATASET,
                ValidationResult.FAILED,
                details=error_details,
            )
            LOGGER.error(dumps({SUCCESS_KEY: False, **error_details}))
            return

        self.process_metadata()
        self.process_assets()

    def process_metadata(self) -> None:
        for index, metadata_file in enumerate(self.dataset_metadata):
            self.processing_assets_model(
                hash_key=self.hash_key,
                range_key=f"{ProcessingAssetType.METADATA.value}{DB_KEY_SEPARATOR}{index}",
                url=metadata_file[PROCESSING_ASSET_URL_KEY],
            ).save()

    def process_assets(self) -> None:
        for index, asset in enumerate(self.dataset_assets):
            self.processing_assets_model(
                hash_key=self.hash_key,
                range_key=f"{ProcessingAssetType.DATA.value}{DB_KEY_SEPARATOR}{index}",
                url=asset[PROCESSING_ASSET_URL_KEY],
                multihash=asset[PROCESSING_ASSET_MULTIHASH_KEY],
            ).save()

    def validate(self, url: str) -> None:  # pylint: disable=too-complex
        self.traversed_urls.append(url)
        object_json = self.get_object(url)

        stac_type = object_json[STAC_TYPE_KEY]
        validator = STAC_TYPE_VALIDATION_MAP[stac_type]

        try:
            validator.validate(object_json)
        except ValidationError as error:
            self.validation_result_factory.save(
                url,
                Check.JSON_SCHEMA,
                ValidationResult.FAILED,
                details={MESSAGE_KEY: str(error)},
            )
            raise
        self.validation_result_factory.save(url, Check.JSON_SCHEMA, ValidationResult.PASSED)
        self.dataset_metadata.append({PROCESSING_ASSET_URL_KEY: url})

        for asset in object_json.get(STAC_ASSETS_KEY, {}).values():
            asset_url = maybe_convert_relative_url_to_absolute(asset[STAC_HREF_KEY], url)

            asset_dict = {
                PROCESSING_ASSET_URL_KEY: asset_url,
                PROCESSING_ASSET_MULTIHASH_KEY: asset[STAC_FILE_CHECKSUM_KEY],
            }
            LOGGER.debug(dumps({PROCESSING_ASSET_ASSET_KEY: asset_dict}))
            self.dataset_assets.append(asset_dict)

        for link_object in object_json[STAC_LINKS_KEY]:
            next_url = maybe_convert_relative_url_to_absolute(link_object[STAC_HREF_KEY], url)

            if next_url not in self.traversed_urls:
                self.validate(next_url)

    def get_object(self, url: str) -> JsonObject:
        try:
            url_stream = self.url_reader(url)
        except ClientError as error:
            self.validation_result_factory.save(
                url,
                Check.STAGING_ACCESS,
                ValidationResult.FAILED,
                details={MESSAGE_KEY: str(error)},
            )
            raise
        try:
            json_object: JsonObject = load(
                url_stream, object_pairs_hook=self.duplicate_object_names_report_builder(url)
            )
        except JSONDecodeError as error:
            self.validation_result_factory.save(
                url, Check.JSON_PARSE, ValidationResult.FAILED, details={MESSAGE_KEY: str(error)}
            )
            raise
        return json_object

    def duplicate_object_names_report_builder(
        self, url: str
    ) -> Callable[[List[Tuple[str, Any]]], JsonObject]:
        def report_duplicate_object_names(object_pairs: List[Tuple[str, Any]]) -> JsonObject:
            result = {}
            for key, value in object_pairs:
                if key in result:
                    self.validation_result_factory.save(
                        url,
                        Check.DUPLICATE_OBJECT_KEY,
                        ValidationResult.FAILED,
                        details={MESSAGE_KEY: f"Found duplicate object name “{key}” in “{url}”"},
                    )
                else:
                    result[key] = value
            return result

        return report_duplicate_object_names



"""
Dataset-versions endpoint Lambda function.
"""
from typing import Callable, MutableMapping

from ..api_responses import handle_request
from ..types import JsonObject
from .create import create_dataset_version

REQUEST_HANDLERS: MutableMapping[str, Callable[[JsonObject], JsonObject]] = {
    "POST": create_dataset_version,
}


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    return handle_request(event, REQUEST_HANDLERS)



from enum import Enum
from json import dumps, loads
from typing import TYPE_CHECKING, Optional

import boto3
from linz_logger import get_log

from .api_keys import SUCCESS_KEY
from .boto3_config import CONFIG
from .import_file_batch_job_id_keys import ASSET_JOB_ID_KEY, METADATA_JOB_ID_KEY
from .models import DATASET_ID_PREFIX, DB_KEY_SEPARATOR, VERSION_ID_PREFIX
from .step_function_keys import (
    ASSET_UPLOAD_KEY,
    DATASET_ID_KEY,
    ERRORS_KEY,
    ERROR_CHECK_KEY,
    ERROR_DETAILS_KEY,
    ERROR_RESULT_KEY,
    ERROR_URL_KEY,
    FAILED_TASKS_KEY,
    FAILURE_REASONS_KEY,
    IMPORT_DATASET_KEY,
    JOB_STATUS_RUNNING,
    JOB_STATUS_SUCCEEDED,
    METADATA_UPLOAD_KEY,
    S3_BATCH_RESPONSE_KEY,
    S3_BATCH_STATUS_FAILED,
    STATUS_KEY,
    STEP_FUNCTION_KEY,
    VALIDATION_KEY,
    VERSION_ID_KEY,
)
from .sts import get_account_number
from .types import JsonList, JsonObject
from .validation_results_model import ValidationResult, validation_results_model_with_meta

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_s3control import S3ControlClient
    from mypy_boto3_stepfunctions import SFNClient
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3ControlClient = SFNClient = object  # pragma: no mutate


class Outcome(Enum):
    PASSED = "Passed"
    PENDING = "Pending"
    FAILED = "Failed"
    SKIPPED = "Skipped"


SUCCESS_TO_VALIDATION_OUTCOME_MAPPING = {
    True: Outcome.PASSED,
    False: Outcome.FAILED,
    None: Outcome.PENDING,
}


STEP_FUNCTIONS_CLIENT: SFNClient = boto3.client("stepfunctions", config=CONFIG)
S3CONTROL_CLIENT: S3ControlClient = boto3.client("s3control", config=CONFIG)
LOGGER = get_log()


def get_tasks_status(
    step_function_status: str,
    dataset_id: str,
    version_id: str,
    validation_success: Optional[bool],
    import_dataset_jobs: JsonObject,
) -> JsonObject:
    validation_errors = get_step_function_validation_results(dataset_id, version_id)
    validation_outcome = get_validation_outcome(
        step_function_status, validation_errors, validation_success
    )

    metadata_upload_status = get_import_job_status(import_dataset_jobs, METADATA_JOB_ID_KEY)
    asset_upload_status = get_import_job_status(import_dataset_jobs, ASSET_JOB_ID_KEY)

    # Failed validation implies uploads will never happen
    if (
        metadata_upload_status[STATUS_KEY] == Outcome.PENDING.value
        and asset_upload_status[STATUS_KEY] == Outcome.PENDING.value
        and validation_outcome in [Outcome.FAILED, Outcome.SKIPPED]
    ):
        metadata_upload_status[STATUS_KEY] = asset_upload_status[STATUS_KEY] = Outcome.SKIPPED.value

    return {
        VALIDATION_KEY: {STATUS_KEY: validation_outcome.value, ERRORS_KEY: validation_errors},
        METADATA_UPLOAD_KEY: metadata_upload_status,
        ASSET_UPLOAD_KEY: asset_upload_status,
    }


def get_import_status_given_arn(execution_arn_key: str) -> JsonObject:
    step_function_resp = STEP_FUNCTIONS_CLIENT.describe_execution(executionArn=execution_arn_key)
    assert "status" in step_function_resp, step_function_resp
    LOGGER.debug(dumps({"step function response": step_function_resp}, default=str))

    step_function_input = loads(step_function_resp["input"])
    step_function_output = loads(step_function_resp.get("output", "{}"))
    step_function_status = step_function_resp["status"]

    dataset_id = step_function_input[DATASET_ID_KEY]
    version_id = step_function_input[VERSION_ID_KEY]
    validation_success = step_function_output.get(VALIDATION_KEY, {}).get(SUCCESS_KEY)
    import_dataset_jobs = step_function_output.get(IMPORT_DATASET_KEY, {})

    tasks_status = get_tasks_status(
        step_function_status, dataset_id, version_id, validation_success, import_dataset_jobs
    )
    return {STEP_FUNCTION_KEY: {"status": step_function_status.title()}, **tasks_status}


def get_validation_outcome(
    step_function_status: str, validation_errors: JsonList, validation_success: Optional[bool]
) -> Outcome:
    validation_status = SUCCESS_TO_VALIDATION_OUTCOME_MAPPING[validation_success]
    if validation_status == Outcome.PENDING:
        # Some statuses are not reported by the step function
        if validation_errors:
            validation_status = Outcome.FAILED
        elif step_function_status not in [JOB_STATUS_RUNNING, JOB_STATUS_SUCCEEDED]:
            validation_status = Outcome.SKIPPED
    return validation_status


def get_import_job_status(import_dataset_jobs: JsonObject, job_id_key: str) -> JsonObject:
    if s3_job_id := import_dataset_jobs.get(job_id_key):
        return get_s3_batch_copy_status(s3_job_id)
    return {STATUS_KEY: Outcome.PENDING.value, ERRORS_KEY: []}


def get_step_function_validation_results(dataset_id: str, version_id: str) -> JsonList:
    hash_key = get_hash_key(dataset_id, version_id)

    errors = []
    validation_results_model = validation_results_model_with_meta()
    for (
        validation_item
    ) in validation_results_model.validation_outcome_index.query(  # pylint: disable=no-member
        hash_key=hash_key,
        range_key_condition=validation_results_model.result == ValidationResult.FAILED.value,
    ):
        _, check_type, _, url = validation_item.sk.split("#", maxsplit=4)
        errors.append(
            {
                ERROR_CHECK_KEY: check_type,
                ERROR_RESULT_KEY: validation_item.result,
                ERROR_URL_KEY: url,
                ERROR_DETAILS_KEY: validation_item.details.attribute_values,
            }
        )

    return errors


def get_s3_batch_copy_status(s3_batch_copy_job_id: str) -> JsonObject:
    account_number = get_account_number()
    s3_batch_copy_resp = S3CONTROL_CLIENT.describe_job(
        AccountId=account_number,
        JobId=s3_batch_copy_job_id,
    )
    assert "Job" in s3_batch_copy_resp, s3_batch_copy_resp
    LOGGER.debug(dumps({S3_BATCH_RESPONSE_KEY: s3_batch_copy_resp}, default=str))

    s3_batch_copy_status = s3_batch_copy_resp["Job"]["Status"]
    failure_reasons = s3_batch_copy_resp["Job"]["FailureReasons"]
    failed_tasks = s3_batch_copy_resp["Job"]["ProgressSummary"]["NumberOfTasksFailed"]

    if failed_tasks > 0:
        s3_batch_copy_status = S3_BATCH_STATUS_FAILED

    return {
        STATUS_KEY: s3_batch_copy_status,
        ERRORS_KEY: {FAILED_TASKS_KEY: failed_tasks, FAILURE_REASONS_KEY: failure_reasons},
    }


def get_hash_key(dataset_id: str, dataset_version_id: str) -> str:
    return (
        f"{DATASET_ID_PREFIX}{dataset_id}{DB_KEY_SEPARATOR}{VERSION_ID_PREFIX}{dataset_version_id}"
    )



from enum import Enum

from .environment import PRODUCTION_ENVIRONMENT_NAME, environment_name


def prefix_non_prod_name(name: str) -> str:
    env_name = environment_name()
    if env_name == PRODUCTION_ENVIRONMENT_NAME:
        return name

    return f"{env_name}-{name}"


class Resource(Enum):
    @property
    def resource_name(self) -> str:
        return prefix_non_prod_name(self.value)

    API_USERS_ROLE_NAME = "api-users"
    CLOUDWATCH_RULE_NAME = "geostore-cloudwatch-rule"
    DATASETS_ENDPOINT_FUNCTION_NAME = "datasets"
    DATASET_VERSIONS_ENDPOINT_FUNCTION_NAME = "dataset-versions"
    IMPORT_STATUS_ENDPOINT_FUNCTION_NAME = "import-status"
    S3_USERS_ROLE_NAME = "s3-users"
    STAGING_USERS_ROLE_NAME = "staging-users"
    STAGING_BUCKET_NAME = "linz-geostore-staging"
    STORAGE_BUCKET_NAME = "linz-geostore"
    SNS_TOPIC_NAME = "geostore-import-status"



"""Delete dataset function."""
from http import HTTPStatus
from typing import TYPE_CHECKING

import boto3
from jsonschema import ValidationError, validate
from pynamodb.exceptions import DoesNotExist

from ..api_responses import error_response, success_response
from ..boto3_config import CONFIG
from ..datasets_model import datasets_model_with_meta
from ..models import DATASET_ID_PREFIX
from ..resources import Resource
from ..step_function_keys import DATASET_ID_SHORT_KEY
from ..types import JsonObject

if TYPE_CHECKING:
    # When type checking we want to use the third party package's stub
    from mypy_boto3_s3 import S3Client
else:
    # In production we want to avoid depending on a package which has no runtime impact
    S3Client = object  # pragma: no mutate

S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)


def delete_dataset(body: JsonObject) -> JsonObject:
    """DELETE: Delete Dataset."""

    body_schema = {
        "type": "object",
        "properties": {DATASET_ID_SHORT_KEY: {"type": "string"}},
        "required": [DATASET_ID_SHORT_KEY],
    }

    # request body validation
    try:
        validate(body, body_schema)
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    datasets_model_class = datasets_model_with_meta()

    # get dataset to delete
    dataset_id = body[DATASET_ID_SHORT_KEY]
    try:
        dataset = datasets_model_class.get(
            hash_key=f"{DATASET_ID_PREFIX}{dataset_id}", consistent_read=True
        )
    except DoesNotExist:
        return error_response(HTTPStatus.NOT_FOUND, f"dataset '{dataset_id}' does not exist")

    # Verify that the dataset is empty
    list_objects_response = S3_CLIENT.list_objects_v2(
        Bucket=Resource.STORAGE_BUCKET_NAME.resource_name, MaxKeys=1, Prefix=f"{dataset_id}/"
    )
    if list_objects_response["KeyCount"]:
        return error_response(
            HTTPStatus.CONFLICT,
            f"Can’t delete dataset “{dataset_id}”: dataset versions still exist",
        )

    # delete dataset
    dataset.delete()

    return success_response(HTTPStatus.NO_CONTENT, {})



"""Import Status handler function."""
from http import HTTPStatus
from json import dumps

from jsonschema import ValidationError, validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY
from ..api_responses import error_response, success_response
from ..error_response_keys import ERROR_KEY
from ..step_function import get_import_status_given_arn
from ..step_function_keys import EXECUTION_ARN_KEY
from ..types import JsonObject

LOGGER = get_log()


def get_import_status(body: JsonObject) -> JsonObject:
    LOGGER.debug(dumps({EVENT_KEY: body}))

    try:
        validate(
            body,
            {
                "type": "object",
                "properties": {EXECUTION_ARN_KEY: {"type": "string"}},
                "required": [EXECUTION_ARN_KEY],
            },
        )
    except ValidationError as err:
        LOGGER.warning(dumps({ERROR_KEY: err}, default=str))
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    response_body = get_import_status_given_arn(body[EXECUTION_ARN_KEY])

    return success_response(HTTPStatus.OK, response_body)



from json import dumps, load
from os.path import basename
from typing import TYPE_CHECKING, Dict, Iterable

import boto3
from linz_logger import get_log

from ..boto3_config import CONFIG
from ..import_dataset_file import get_import_result
from ..stac_format import STAC_ASSETS_KEY, STAC_HREF_KEY, STAC_LINKS_KEY
from ..types import JsonObject

S3_BODY_KEY = "Body"

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3.type_defs import PutObjectOutputTypeDef
else:
    PutObjectOutputTypeDef = JsonObject  # pragma: no mutate
    S3Client = object  # pragma: no mutate

TARGET_S3_CLIENT: S3Client = boto3.client("s3", config=CONFIG)
LOGGER = get_log()


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    return get_import_result(event, importer)


def importer(
    source_bucket_name: str,
    original_key: str,
    target_bucket_name: str,
    new_key: str,
    source_s3_client: S3Client,
) -> PutObjectOutputTypeDef:
    get_object_response = source_s3_client.get_object(Bucket=source_bucket_name, Key=original_key)
    assert S3_BODY_KEY in get_object_response, get_object_response

    metadata = load(get_object_response["Body"])

    assets = metadata.get(STAC_ASSETS_KEY, {}).values()
    change_href_to_basename(assets)

    links = metadata.get(STAC_LINKS_KEY, [])
    change_href_to_basename(links)

    return TARGET_S3_CLIENT.put_object(
        Bucket=target_bucket_name,
        Key=new_key,
        Body=dumps(metadata).encode(),
    )


def change_href_to_basename(items: Iterable[Dict[str, str]]) -> None:
    for item in items:
        item[STAC_HREF_KEY] = basename(item[STAC_HREF_KEY])



STAC_ASSETS_KEY = "assets"
STAC_DESCRIPTION_KEY = "description"
STAC_EXTENSIONS_KEY = "stac_extensions"
STAC_EXTENT_BBOX_KEY = "bbox"
STAC_EXTENT_KEY = "extent"
STAC_EXTENT_SPATIAL_KEY = "spatial"
STAC_EXTENT_TEMPORAL_INTERVAL_KEY = "interval"
STAC_EXTENT_TEMPORAL_KEY = "temporal"
STAC_FILE_CHECKSUM_KEY = "file:checksum"
STAC_GEOMETRY_KEY = "geometry"
STAC_HREF_KEY = "href"
STAC_ID_KEY = "id"
STAC_LICENSE_KEY = "license"
STAC_LINKS_KEY = "links"
STAC_MEDIA_TYPE_GEOJSON = "application/geo+json"
STAC_MEDIA_TYPE_JSON = "application/json"
STAC_PROPERTIES_DATETIME_KEY = "datetime"
STAC_PROPERTIES_KEY = "properties"
STAC_REL_CHILD = "child"
STAC_REL_ITEM = "item"
STAC_REL_KEY = "rel"
STAC_REL_PARENT = "parent"
STAC_REL_ROOT = "root"
STAC_REL_SELF = "self"
STAC_TITLE_KEY = "title"
STAC_TYPE_CATALOG = "Catalog"
STAC_TYPE_COLLECTION = "Collection"
STAC_TYPE_ITEM = "Feature"
STAC_TYPE_KEY = "type"
STAC_VERSION = "1.0.0"
STAC_VERSION_KEY = "stac_version"

LINZ_SCHEMA_DIRECTORY = "stac/v0.0.7/linz"
LINZ_SCHEMA_PATH = f"{LINZ_SCHEMA_DIRECTORY}/schema.json"
LINZ_STAC_EXTENSION_URL = f"https://linz.github.io/{LINZ_SCHEMA_PATH}"
LINZ_STAC_CREATED_KEY = "linz:created"
LINZ_STAC_SECURITY_CLASSIFICATION_KEY = "linz:security_classification"
LINZ_STAC_LIFECYCLE_KEY = "linz:lifecycle"
LINZ_STAC_PROVIDERS_KEY = "linz:providers"
LINZ_STAC_UPDATED_KEY = "linz:updated"

QUALITY_SCHEMA_PATH = "stac/v0.0.7/quality/schema.json"
QUALITY_LINEAGE_KEY = "quality:lineage"

VERSION_VERSION_KEY = "version"



DB_KEY_SEPARATOR = "#"

CHECK_ID_PREFIX = f"CHECK{DB_KEY_SEPARATOR}"
DATASET_ID_PREFIX = f"DATASET{DB_KEY_SEPARATOR}"
URL_ID_PREFIX = f"URL{DB_KEY_SEPARATOR}"
VERSION_ID_PREFIX = f"VERSION{DB_KEY_SEPARATOR}"



from functools import cached_property
from json import load
from os.path import dirname, join

from jsonschema import Draft7Validator, FormatChecker, RefResolver
from jsonschema._utils import URIDict
from jsonschema.validators import extend

from ..stac_format import LINZ_SCHEMA_PATH, QUALITY_SCHEMA_PATH
from ..types import JsonObject

STAC_SPEC_PATH = "stac-spec/v1.0.0"
STAC_ITEM_SPEC_PATH = f"{STAC_SPEC_PATH}/item-spec/json-schema"

PROJECTION_STAC_SCHEMA_PATH = "projection/v1.0.0/schema.json"
VERSION_STAC_SCHEMA_PATH = "version/v1.0.0/schema.json"


class Schema:
    def __init__(self, path: str):
        self.path = path

    @cached_property
    def as_dict(self) -> JsonObject:
        with open(join(dirname(__file__), self.path), encoding="utf-8") as file_pointer:
            result: JsonObject = load(file_pointer)
            return result

    @cached_property
    def schema_id(self) -> str:
        id_: str = self.as_dict["$id"]
        return id_

    @cached_property
    def uri(self) -> str:
        uri_: str = URIDict().normalize(self.schema_id)
        return uri_


CATALOG_SCHEMA = Schema(f"{STAC_SPEC_PATH}/catalog-spec/json-schema/catalog.json")
COLLECTION_SCHEMA = Schema(LINZ_SCHEMA_PATH)
ITEM_SCHEMA = Schema(f"{STAC_ITEM_SPEC_PATH}/item.json")

schema_store = {}
for schema in [
    CATALOG_SCHEMA,
    COLLECTION_SCHEMA,
    ITEM_SCHEMA,
    Schema("geojson-spec/Feature.json"),
    Schema("geojson-spec/Geometry.json"),
    Schema(PROJECTION_STAC_SCHEMA_PATH),
    Schema(f"{STAC_SPEC_PATH}/collection-spec/json-schema/collection.json"),
    Schema(f"{STAC_ITEM_SPEC_PATH}/basics.json"),
    Schema(f"{STAC_ITEM_SPEC_PATH}/datetime.json"),
    Schema(f"{STAC_ITEM_SPEC_PATH}/instrument.json"),
    Schema(f"{STAC_ITEM_SPEC_PATH}/licensing.json"),
    Schema(f"{STAC_ITEM_SPEC_PATH}/provider.json"),
    Schema(VERSION_STAC_SCHEMA_PATH),
    Schema(QUALITY_SCHEMA_PATH),
]:
    # Normalize URLs the same way as jsonschema does
    schema_store[schema.uri] = schema.as_dict

BaseSTACValidator = extend(Draft7Validator)
BaseSTACValidator.format_checker = FormatChecker()

STACCatalogSchemaValidator = extend(BaseSTACValidator)(
    resolver=RefResolver.from_schema(CATALOG_SCHEMA.as_dict, store=schema_store),
    schema=CATALOG_SCHEMA.as_dict,
)

STACCollectionSchemaValidator = extend(BaseSTACValidator)(
    resolver=RefResolver.from_schema(COLLECTION_SCHEMA.as_dict, store=schema_store),
    schema=COLLECTION_SCHEMA.as_dict,
)

STACItemSchemaValidator = extend(BaseSTACValidator)(
    resolver=RefResolver.from_schema(ITEM_SCHEMA.as_dict, store=schema_store),
    schema=ITEM_SCHEMA.as_dict,
)



from os import environ

ENV_NAME_VARIABLE_NAME = "GEOSTORE_ENV_NAME"
PRODUCTION_ENVIRONMENT_NAME = "prod"


def environment_name() -> str:
    return environ.get(ENV_NAME_VARIABLE_NAME, PRODUCTION_ENVIRONMENT_NAME)



"""Get datasets functions."""
from http import HTTPStatus

from jsonschema import ValidationError, validate
from pynamodb.exceptions import DoesNotExist

from ..api_responses import error_response, success_response
from ..datasets_model import datasets_model_with_meta
from ..models import DATASET_ID_PREFIX
from ..step_function_keys import DATASET_ID_SHORT_KEY, TITLE_KEY
from ..types import JsonObject
from .list import list_datasets


def handle_get(body: JsonObject) -> JsonObject:
    if DATASET_ID_SHORT_KEY in body:
        return get_dataset_single(body)

    if TITLE_KEY in body:
        return get_dataset_filter(body)

    if body == {}:
        return list_datasets()

    return error_response(HTTPStatus.BAD_REQUEST, "Unhandled request")


def get_dataset_single(body: JsonObject) -> JsonObject:
    """GET: Get single Dataset."""

    body_schema = {
        "type": "object",
        "properties": {DATASET_ID_SHORT_KEY: {"type": "string"}},
        "required": [DATASET_ID_SHORT_KEY],
    }

    # request body validation
    try:
        validate(body, body_schema)
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    datasets_model_class = datasets_model_with_meta()

    # get dataset
    try:
        dataset = datasets_model_class.get(
            hash_key=f"{DATASET_ID_PREFIX}{body[DATASET_ID_SHORT_KEY]}", consistent_read=True
        )
    except DoesNotExist:
        return error_response(
            HTTPStatus.NOT_FOUND, f"dataset '{body[DATASET_ID_SHORT_KEY]}' does not exist"
        )

    # return response
    resp_body = dataset.as_dict()

    return success_response(HTTPStatus.OK, resp_body)


def get_dataset_filter(body: JsonObject) -> JsonObject:
    """GET: Get Datasets by filter."""

    body_schema = {
        "type": "object",
        "properties": {TITLE_KEY: {"type": "string"}},
        "required": [TITLE_KEY],
    }

    # request body validation
    try:
        validate(body, body_schema)
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    # dataset query by filter
    datasets_model_class = datasets_model_with_meta()
    datasets = datasets_model_class.datasets_title_idx.query(hash_key=body[TITLE_KEY])

    # return response
    resp_body = []
    for dataset in datasets:
        resp_item = dataset.as_dict()
        resp_body.append(resp_item)

    return success_response(HTTPStatus.OK, resp_body)



from http import HTTPStatus
from http.client import responses as http_responses
from typing import Callable, Mapping, Union

from jsonschema import ValidationError, validate

from .api_keys import MESSAGE_KEY
from .aws_keys import BODY_KEY, HTTP_METHOD_KEY, STATUS_CODE_KEY
from .types import JsonList, JsonObject


def error_response(code: int, message: str) -> JsonObject:
    return {STATUS_CODE_KEY: code, BODY_KEY: {MESSAGE_KEY: f"{http_responses[code]}: {message}"}}


def success_response(code: int, body: Union[JsonList, JsonObject]) -> JsonObject:
    return {STATUS_CODE_KEY: code, BODY_KEY: body}


def handle_request(
    event: JsonObject, request_handlers: Mapping[str, Callable[[JsonObject], JsonObject]]
) -> JsonObject:
    """Main Lambda entry point."""

    # request validation
    try:
        validate(
            event,
            {
                "type": "object",
                "properties": {
                    HTTP_METHOD_KEY: {"type": "string", "enum": list(request_handlers.keys())},
                    BODY_KEY: {"type": "object"},
                },
                "required": [HTTP_METHOD_KEY, BODY_KEY],
            },
        )
    except ValidationError as err:
        return error_response(HTTPStatus.BAD_REQUEST, err.message)

    method = event[HTTP_METHOD_KEY]
    return request_handlers[method](event[BODY_KEY])



from json import dumps

from jsonschema import ValidationError, validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY, SUCCESS_KEY
from ..error_response_keys import ERROR_MESSAGE_KEY
from ..models import DATASET_ID_PREFIX, DB_KEY_SEPARATOR, VERSION_ID_PREFIX
from ..step_function_keys import DATASET_ID_KEY, VERSION_ID_KEY
from ..types import JsonObject
from ..validation_results_model import ValidationResult, validation_results_model_with_meta

LOGGER = get_log()


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    LOGGER.debug(dumps({EVENT_KEY: event}))

    try:
        validate(
            event,
            {
                "type": "object",
                "properties": {
                    DATASET_ID_KEY: {"type": "string"},
                    VERSION_ID_KEY: {"type": "string"},
                },
                "required": [DATASET_ID_KEY, VERSION_ID_KEY],
            },
        )
    except ValidationError as error:
        return {ERROR_MESSAGE_KEY: error.message}

    validation_results_model = validation_results_model_with_meta()
    success = not bool(
        validation_results_model.validation_outcome_index.count(  # pylint: disable=no-member
            (
                f"{DATASET_ID_PREFIX}{event['dataset_id']}"
                f"{DB_KEY_SEPARATOR}{VERSION_ID_PREFIX}{event['version_id']}"
            ),
            range_key_condition=validation_results_model.result == ValidationResult.FAILED.value,
            limit=1,
        )
    )

    result = {SUCCESS_KEY: success}
    LOGGER.debug(dumps(result))
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['geostore',
 'geostore.check_files_checksums',
 'geostore.check_stac_metadata',
 'geostore.content_iterator',
 'geostore.dataset_versions',
 'geostore.datasets',
 'geostore.import_asset_file',
 'geostore.import_dataset',
 'geostore.import_metadata_file',
 'geostore.import_status',
 'geostore.notify_status_update',
 'geostore.populate_catalog',
 'geostore.update_dataset_catalog',
 'geostore.upload_status',
 'geostore.validation_summary']

package_data = \
{'': ['*'],
 'geostore.check_stac_metadata': ['file/*',
                                  'file/v1.0.0/*',
                                  'file/v2.0.0/*',
                                  'geojson-spec/*',
                                  'projection/*',
                                  'projection/v1.0.0/*',
                                  'stac-spec/*',
                                  'stac-spec/dev/catalog-spec/json-schema/*',
                                  'stac-spec/dev/collection-spec/json-schema/*',
                                  'stac-spec/dev/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/checksum/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/collection-assets/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/datacube/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/eo/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/item-assets/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/label/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/pointcloud/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/projection/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/sar/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/sat/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/scientific/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/single-file-stac/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/tiled-assets/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/timestamps/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/version/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/extensions/view/json-schema/*',
                                  'stac-spec/v1.0.0-beta.2/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/extensions/eo/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/extensions/projection/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/extensions/scientific/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/extensions/view/json-schema/*',
                                  'stac-spec/v1.0.0-rc.1/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.2/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.2/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.2/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.3/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.3/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.3/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.4/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.4/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0-rc.4/item-spec/json-schema/*',
                                  'stac-spec/v1.0.0/catalog-spec/json-schema/*',
                                  'stac-spec/v1.0.0/collection-spec/json-schema/*',
                                  'stac-spec/v1.0.0/item-spec/json-schema/*',
                                  'stac/*',
                                  'stac/v0.0.10/aerial-photo/*',
                                  'stac/v0.0.10/aerial-photo/examples/*',
                                  'stac/v0.0.10/aerial-photo/tests/*',
                                  'stac/v0.0.10/camera/*',
                                  'stac/v0.0.10/camera/examples/*',
                                  'stac/v0.0.10/camera/tests/*',
                                  'stac/v0.0.10/film/*',
                                  'stac/v0.0.10/film/examples/*',
                                  'stac/v0.0.10/film/tests/*',
                                  'stac/v0.0.10/historical-imagery/*',
                                  'stac/v0.0.10/historical-imagery/examples/*',
                                  'stac/v0.0.10/historical-imagery/tests/*',
                                  'stac/v0.0.10/linz/*',
                                  'stac/v0.0.10/linz/examples/*',
                                  'stac/v0.0.10/linz/tests/*',
                                  'stac/v0.0.10/quality/*',
                                  'stac/v0.0.10/quality/examples/*',
                                  'stac/v0.0.10/quality/tests/*',
                                  'stac/v0.0.10/scanning/*',
                                  'stac/v0.0.10/scanning/examples/*',
                                  'stac/v0.0.10/template/*',
                                  'stac/v0.0.10/template/examples/*',
                                  'stac/v0.0.10/template/tests/*',
                                  'stac/v0.0.2/camera/*',
                                  'stac/v0.0.2/camera/examples/*',
                                  'stac/v0.0.2/historical_imagery/*',
                                  'stac/v0.0.2/linz/*',
                                  'stac/v0.0.2/linz/examples/*',
                                  'stac/v0.0.2/quality/*',
                                  'stac/v0.0.2/quality/examples/*',
                                  'stac/v0.0.2/template/*',
                                  'stac/v0.0.2/template/examples/*',
                                  'stac/v0.0.3/camera/*',
                                  'stac/v0.0.3/camera/examples/*',
                                  'stac/v0.0.3/historical_imagery/*',
                                  'stac/v0.0.3/linz/*',
                                  'stac/v0.0.3/linz/examples/*',
                                  'stac/v0.0.3/quality/*',
                                  'stac/v0.0.3/quality/examples/*',
                                  'stac/v0.0.3/template/*',
                                  'stac/v0.0.3/template/examples/*',
                                  'stac/v0.0.4/camera/*',
                                  'stac/v0.0.4/camera/examples/*',
                                  'stac/v0.0.4/historical_imagery/*',
                                  'stac/v0.0.4/linz/*',
                                  'stac/v0.0.4/linz/examples/*',
                                  'stac/v0.0.4/quality/*',
                                  'stac/v0.0.4/quality/examples/*',
                                  'stac/v0.0.4/template/*',
                                  'stac/v0.0.4/template/examples/*',
                                  'stac/v0.0.5/aerial-photo/*',
                                  'stac/v0.0.5/aerial-photo/examples/*',
                                  'stac/v0.0.5/aerial-photo/non-examples/*',
                                  'stac/v0.0.5/camera/*',
                                  'stac/v0.0.5/camera/examples/*',
                                  'stac/v0.0.5/camera/non-examples/*',
                                  'stac/v0.0.5/film/*',
                                  'stac/v0.0.5/film/examples/*',
                                  'stac/v0.0.5/film/non-examples/*',
                                  'stac/v0.0.5/historical_imagery/*',
                                  'stac/v0.0.5/linz/*',
                                  'stac/v0.0.5/linz/examples/*',
                                  'stac/v0.0.5/linz/non-examples/*',
                                  'stac/v0.0.5/quality/*',
                                  'stac/v0.0.5/quality/examples/*',
                                  'stac/v0.0.5/quality/non-examples/*',
                                  'stac/v0.0.5/scanning/*',
                                  'stac/v0.0.5/scanning/examples/*',
                                  'stac/v0.0.5/template/*',
                                  'stac/v0.0.5/template/examples/*',
                                  'stac/v0.0.5/template/non-examples/*',
                                  'stac/v0.0.6/aerial-photo/*',
                                  'stac/v0.0.6/aerial-photo/examples/*',
                                  'stac/v0.0.6/aerial-photo/non-examples/*',
                                  'stac/v0.0.6/camera/*',
                                  'stac/v0.0.6/camera/examples/*',
                                  'stac/v0.0.6/camera/non-examples/*',
                                  'stac/v0.0.6/film/*',
                                  'stac/v0.0.6/film/examples/*',
                                  'stac/v0.0.6/film/non-examples/*',
                                  'stac/v0.0.6/historical_imagery/*',
                                  'stac/v0.0.6/linz/*',
                                  'stac/v0.0.6/linz/examples/*',
                                  'stac/v0.0.6/linz/non-examples/*',
                                  'stac/v0.0.6/quality/*',
                                  'stac/v0.0.6/quality/examples/*',
                                  'stac/v0.0.6/quality/non-examples/*',
                                  'stac/v0.0.6/scanning/*',
                                  'stac/v0.0.6/scanning/examples/*',
                                  'stac/v0.0.6/template/*',
                                  'stac/v0.0.6/template/examples/*',
                                  'stac/v0.0.6/template/non-examples/*',
                                  'stac/v0.0.7/aerial-photo/*',
                                  'stac/v0.0.7/aerial-photo/examples/*',
                                  'stac/v0.0.7/aerial-photo/non-examples/*',
                                  'stac/v0.0.7/camera/*',
                                  'stac/v0.0.7/camera/examples/*',
                                  'stac/v0.0.7/camera/tests/*',
                                  'stac/v0.0.7/film/*',
                                  'stac/v0.0.7/film/examples/*',
                                  'stac/v0.0.7/film/non-examples/*',
                                  'stac/v0.0.7/historical_imagery/*',
                                  'stac/v0.0.7/linz/*',
                                  'stac/v0.0.7/linz/examples/*',
                                  'stac/v0.0.7/linz/non-examples/*',
                                  'stac/v0.0.7/quality/*',
                                  'stac/v0.0.7/quality/examples/*',
                                  'stac/v0.0.7/quality/non-examples/*',
                                  'stac/v0.0.7/scanning/*',
                                  'stac/v0.0.7/scanning/examples/*',
                                  'stac/v0.0.7/template/*',
                                  'stac/v0.0.7/template/examples/*',
                                  'stac/v0.0.7/template/non-examples/*',
                                  'stac/v0.0.8/*',
                                  'stac/v0.0.8/aerial-photo/*',
                                  'stac/v0.0.8/aerial-photo/examples/*',
                                  'stac/v0.0.8/aerial-photo/tests/*',
                                  'stac/v0.0.8/camera/*',
                                  'stac/v0.0.8/camera/examples/*',
                                  'stac/v0.0.8/camera/tests/*',
                                  'stac/v0.0.8/film/*',
                                  'stac/v0.0.8/film/examples/*',
                                  'stac/v0.0.8/film/tests/*',
                                  'stac/v0.0.8/historical-imagery/*',
                                  'stac/v0.0.8/linz/*',
                                  'stac/v0.0.8/linz/examples/*',
                                  'stac/v0.0.8/linz/tests/*',
                                  'stac/v0.0.8/quality/*',
                                  'stac/v0.0.8/quality/examples/*',
                                  'stac/v0.0.8/quality/tests/*',
                                  'stac/v0.0.8/scanning/*',
                                  'stac/v0.0.8/scanning/examples/*',
                                  'stac/v0.0.8/template/*',
                                  'stac/v0.0.8/template/examples/*',
                                  'stac/v0.0.8/template/tests/*',
                                  'stac/v0.0.9/*',
                                  'stac/v0.0.9/aerial-photo/*',
                                  'stac/v0.0.9/aerial-photo/examples/*',
                                  'stac/v0.0.9/aerial-photo/tests/*',
                                  'stac/v0.0.9/camera/*',
                                  'stac/v0.0.9/camera/examples/*',
                                  'stac/v0.0.9/camera/tests/*',
                                  'stac/v0.0.9/film/*',
                                  'stac/v0.0.9/film/examples/*',
                                  'stac/v0.0.9/film/tests/*',
                                  'stac/v0.0.9/historical-imagery/*',
                                  'stac/v0.0.9/linz/*',
                                  'stac/v0.0.9/linz/examples/*',
                                  'stac/v0.0.9/linz/tests/*',
                                  'stac/v0.0.9/quality/*',
                                  'stac/v0.0.9/quality/examples/*',
                                  'stac/v0.0.9/quality/tests/*',
                                  'stac/v0.0.9/scanning/*',
                                  'stac/v0.0.9/scanning/examples/*',
                                  'stac/v0.0.9/template/*',
                                  'stac/v0.0.9/template/examples/*',
                                  'stac/v0.0.9/template/tests/*',
                                  'version/*',
                                  'version/v1.0.0/*']}

install_requires = \
['boto3', 'typer']

extras_require = \
{'cdk': ['aws-cdk.aws-dynamodb',
         'aws-cdk.aws-ec2',
         'aws-cdk.aws-ecr',
         'aws-cdk.aws-ecr_assets',
         'aws-cdk.aws-ecs',
         'aws-cdk.aws-events',
         'aws-cdk.aws-events-targets',
         'aws-cdk.aws-iam',
         'aws-cdk.aws-lambda',
         'aws-cdk.aws-lambda-event-sources',
         'aws-cdk.aws-lambda-python',
         'aws-cdk.aws-s3',
         'aws-cdk.aws-sns',
         'aws-cdk.aws-stepfunctions',
         'aws-cdk.aws-stepfunctions_tasks',
         'awscli',
         'cattrs'],
 'check_files_checksums': ['multihash', 'pynamodb', 'linz-logger'],
 'check_stac_metadata': ['jsonschema[format]',
                         'pynamodb',
                         'strict-rfc3339',
                         'linz-logger'],
 'content_iterator': ['jsonschema[format]', 'pynamodb', 'linz-logger'],
 'dataset_versions': ['jsonschema[format]',
                      'pynamodb',
                      'ulid-py',
                      'linz-logger'],
 'datasets': ['jsonschema[format]',
              'pynamodb',
              'pystac',
              'ulid-py',
              'linz-logger'],
 'import_asset_file': ['smart-open[s3]', 'linz-logger'],
 'import_dataset': ['jsonschema[format]',
                    'pynamodb',
                    'smart-open[s3]',
                    'ulid-py',
                    'linz-logger'],
 'import_metadata_file': ['linz-logger'],
 'import_status': ['jsonschema[format]', 'pynamodb', 'linz-logger'],
 'notify_status_update': ['jsonschema[format]',
                          'pynamodb',
                          'slack-sdk[webhook,models]',
                          'linz-logger'],
 'populate_catalog': ['jsonschema[format]', 'pystac', 'linz-logger'],
 'update_dataset_catalog': ['jsonschema[format]',
                            'pynamodb',
                            'ulid-py',
                            'linz-logger'],
 'upload_status': ['jsonschema[format]', 'pynamodb', 'linz-logger'],
 'validation_summary': ['jsonschema[format]', 'pynamodb', 'linz-logger']}

entry_points = \
{'console_scripts': ['geostore = geostore.cli:app']}

setup_kwargs = {
    'name': 'geostore',
    'version': '0.1.4',
    'description': 'Central storage, management and access for important geospatial datasets developed by LINZ',
    'long_description': '# Geostore\n\n[![Deploy](https://github.com/linz/geostore/actions/workflows/deploy.yml/badge.svg)](https://github.com/linz/geostore/actions/workflows/deploy.yml)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/linz/geostore.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/linz/geostore/alerts/)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/linz/geostore.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/linz/geostore/context:python)\n[![CodeQL Analysis](https://github.com/linz/geostore/actions/workflows/codeql-analysis.yml/badge.svg)](https://github.com/linz/geostore/actions/workflows/codeql-analysis.yml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Coverage: 100% branches](https://img.shields.io/badge/Coverage-100%25%20branches-brightgreen.svg)](https://pytest.org/)\n[![Dependabot Status](https://badgen.net/badge/Dependabot/enabled?labelColor=2e3a44&color=blue)](https://github.com/linz/geostore/network/updates)\n[![hadolint: passing](https://img.shields.io/badge/hadolint-passing-brightgreen)](https://github.com/hadolint/hadolint)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Kodiak](https://badgen.net/badge/Kodiak/enabled?labelColor=2e3a44&color=F39938)](https://kodiakhq.com/)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![code style: prettier](https://img.shields.io/badge/code_style-prettier-ff69b4.svg)](https://github.com/prettier/prettier)\n[![pylint: passing](https://img.shields.io/badge/pylint-passing-brightgreen)](https://www.pylint.org/)\n[![Python: 3.8](https://img.shields.io/badge/Python-3.8-blue)](https://www.python.org/)\n[![shellcheck: passing](https://img.shields.io/badge/shellcheck-passing-brightgreen)](https://www.shellcheck.net/)\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-yellow.svg)](https://conventionalcommits.org)\n\nLINZ central storage, management and access solution for important geospatial datasets. Developed by\n[Land Information New Zealand](https://github.com/linz).\n\n## Prerequisites\n\n### Geostore VPC\n\nA Geostore VPC must exist in your AWS account before deploying this application. AT LINZ, VPCs are\nmanaged internally by the IT team. If you are deploying this application outside LINZ, you will need\nto create a VPC with the following tags:\n\n- "ApplicationName": "geostore"\n- "ApplicationLayer": "networking"\n\nYou can achieve this by adding the `networking_stack` (`infrastructure/networking_stack.py)` into\n`app.py` before deployment as a dependency of `application_stack`\n(`infrastructure/application_stack.py`).\n\n### Verify infrastructure settings\n\nThis infrastructure by default includes some Toitū Te Whenua-/LINZ-specific parts, controlled by\nsettings in cdk.json. To disable these, simply remove the context entries or set them to `false`.\nThe settings are:\n\n- `enableLDSAccess`: if true, gives LINZ Data Service/Koordinates read access to the storage bucket.\n- `enableOpenTopographyAccess`: if true, gives OpenTopography read access to the storage bucket.\n\n## Development setup\n\nOne-time setup which generally assumes that you\'re in the project directory.\n\n### Common\n\n1. [Install Docker](https://docs.docker.com/engine/install/ubuntu/)\n2. Configure Docker:\n   1. Add yourself to the "docker" group: `sudo usermod --append --groups=docker "$USER"`\n   2. Log out and back in to enable the new group\n\n### Ubuntu\n\n1. Install [`nvm`](https://github.com/nvm-sh/nvm#installing-and-updating):\n   ```bash\n   cd "$(mktemp --directory)"\n   wget https://raw.githubusercontent.com/nvm-sh/nvm/master/install.sh\n   echo \'b674516f001d331c517be63c1baeaf71de6cbb6d68a44112bf2cff39a6bc246a install.sh\' | sha256sum --check && bash install.sh\n   ```\n2. Install [Poetry](https://python-poetry.org/docs/master/#installation):\n   ```bash\n   cd "$(mktemp --directory)"\n   wget https://raw.githubusercontent.com/python-poetry/poetry/master/install-poetry.py\n   echo \'b35d059be6f343ac1f05ae56e8eaaaebb34da8c92424ee00133821d7f11e3a9c install-poetry.py\' | sha256sum --check && python3 install-poetry.py\n   ```\n3. Install [Pyenv](https://github.com/pyenv/pyenv#installation):\n   ```bash\n   sudo apt-get update\n   sudo apt-get install --no-install-recommends build-essential curl libbz2-dev libffi-dev liblzma-dev libncurses5-dev libreadline-dev libsqlite3-dev libssl-dev libxml2-dev libxmlsec1-dev llvm make tk-dev wget xz-utils zlib1g-dev\n   cd "$(mktemp --directory)"\n   wget https://github.com/pyenv/pyenv-installer/raw/master/bin/pyenv-installer\n   echo \'3aa49f2b3b77556272a80a01fe44d46733f4862dbbbc956002dc944c428bebd8 pyenv-installer\' | sha256sum --check && bash pyenv-installer\n   ```\n4. Enable the above by adding the following to your `~/.bashrc`:\n\n   ```bash\n   if [[ -e "${HOME}/.local/bin" ]]\n   then\n       PATH="${HOME}/.local/bin:${PATH}"\n   fi\n\n   # nvm <https://github.com/nvm-sh/nvm>\n   if [[ -d "${HOME}/.nvm" ]]\n   then\n       export NVM_DIR="${HOME}/.nvm"\n       # shellcheck source=/dev/null\n       [[ -s "${NVM_DIR}/nvm.sh" ]] && . "${NVM_DIR}/nvm.sh"\n       # shellcheck source=/dev/null\n       [[ -s "${NVM_DIR}/bash_completion" ]] && . "${NVM_DIR}/bash_completion"\n   fi\n\n   # Pyenv <https://github.com/pyenv/pyenv>\n   if [[ -e "${HOME}/.pyenv" ]]\n   then\n       PATH="${HOME}/.pyenv/bin:${PATH}"\n       eval "$(pyenv init --path)"\n       eval "$(pyenv init -)"\n       eval "$(pyenv virtualenv-init -)"\n   fi\n   ```\n\n5. Configure Docker:\n   1. Add yourself to the "docker" group: `sudo usermod --append --groups=docker "$USER"`\n   1. Log out and back in to enable the new group\n6. [Install project Node.js](https://github.com/nvm-sh/nvm#long-term-support): `nvm install`\n7. Run `./reset-dev-env.bash --all` to install packages.\n8. Enable the dev environment: `. activate-dev-env.bash`.\n9. Optional: Enable [Dependabot alerts by email](https://github.com/settings/notifications). (This\n   is optional since it currently can\'t be set per repository or organisation, so it affects any\n   repos where you have access to Dependabot alerts.)\n10. Install [`aws-azure-login`](https://github.com/sportradar/aws-azure-login#installation).\n\nRe-run `./reset-dev-env.bash` when packages change. One easy way to use it pretty much seamlessly is\nto run it before every workday, with a crontab entry like this template:\n\n```crontab\nHOME=\'/home/USERNAME\'\n0 2 * * 1-5 export PATH="${HOME}/.pyenv/shims:${HOME}/.pyenv/bin:${HOME}/.poetry/bin:/root/bin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin:/run/current-system/sw/bin" && cd "PATH_TO_GEOSTORE" && ./reset-dev-env.bash --all\n```\n\nReplace `USERNAME` and `PATH_TO_GEOSTORE` with your values, resulting in something like this:\n\n```crontab\nHOME=\'/home/jdoe\'\n0 2 * * 1-5 export PATH="${HOME}/.pyenv/shims:${HOME}/.pyenv/bin:${HOME}/.poetry/bin:/root/bin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin:/run/current-system/sw/bin" && cd "${HOME}/dev/geostore" && ./reset-dev-env.bash --all\n```\n\nRe-run `. activate-dev-env.bash` in each shell.\n\n### Nix\n\n1. Run `nix-shell`.\n1. Optional: Install and [configure](https://direnv.net/docs/hook.html) `direnv` and\n   `direnv allow .` to load the Nix shell whenever you `cd` into the project.\n\nRestart your `nix-shell` when packages change.\n\nWhen setting up the project SDK point it to `.run/python`, which is a symlink to the latest Nix\nshell Python executable.\n\n### Optional\n\nEnable [Dependabot alerts by email](https://github.com/settings/notifications). (This is optional\nsince it currently can\'t be set per repository or organisation, so it affects any repos where you\nhave access to Dependabot alerts.)\n\n## AWS Infrastructure deployment\n\n1. [Configure AWS](https://confluence.linz.govt.nz/display/GEOD/Login+to+AWS+Service+Accounts+via+Azure+in+Command+Line)\n1. Get AWS credentials (see: https://www.npmjs.com/package/aws-azure-login) for 12 hours:\n\n   ```bash\n   aws-azure-login --no-prompt --profile=<AWS-PROFILE-NAME>\n   ```\n\n1. Environment variables\n\n   - **`GEOSTORE_ENV_NAME`:** set deployment environment. For your personal development stack: set\n     GEOSTORE_ENV_NAME to your username.\n\n     ```bash\n     export GEOSTORE_ENV_NAME="$USER"\n     ```\n\n     Other values used by CI pipelines include: prod, nonprod, ci, dev or any string without spaces.\n     Default: test.\n\n   * **`RESOURCE_REMOVAL_POLICY`:** determines if resources containing user content like Geostore\n     Storage S3 bucket or application database tables will be preserved even if they are removed\n     from stack or stack is deleted. Supported values:\n     - DESTROY: destroy resource when removed from stack or stack is deleted (default)\n     - RETAIN: retain orphaned resource when removed from stack or stack is deleted\n\n   - **`GEOSTORE_SAML_IDENTITY_PROVIDER_ARN`:** SAML identity provider AWS ARN.\n\n1. Bootstrap CDK (only once per profile)\n\n   ```bash\n   cdk --profile=<AWS-PROFILE-NAME> bootstrap aws://unknown-account/ap-southeast-2\n   ```\n\n1. Deploy CDK stack\n\n   ```bash\n   cdk --profile=<AWS-PROFILE-NAME> deploy --all\n   ```\n\n   Once comfortable with CDK you can add `--require-approval=never` above to deploy\n   non-interactively.\n\nIf you `export AWS_PROFILE=<AWS-PROFILE-NAME>` you won\'t need the `--profile=<AWS-PROFILE-NAME>`\narguments above.\n\n## Development\n\n### Adding or updating Python dependencies\n\nTo add a development-only package: `poetry add --dev PACKAGE=\'*\'`\n\nTo add a production package:\n\n1. Install the package using `poetry add --optional PACKAGE=\'*\'`.\n1. Put the package in alphabetical order within the list.\n1. Mention the package in the relevant lists in `[tool.poetry.extras]`.\n\n- Make sure to update packages separately from adding packages. Basically, follow this process\n  before running `poetry add`, and do the equivalent when updating Node.js packages or changing\n  Docker base images:\n\n  1.  Check out a new branch on top of origin/master:\n      `git checkout -b update-python-packages origin/master`.\n  1.  Update the Python packages: `poetry update`. The rest of the steps are only necessary if this\n      step changes poetry.lock. Otherwise you can just change back to the original branch and delete\n      "update-python-packages".\n  1.  Commit, push and create pull request.\n  1.  Check out the branch where you originally wanted to run `poetry add`.\n  1.  Rebase the branch onto the package update branch: `git rebase update-python-packages`.\n\n  At this point any `poetry add` commands should not result in any package updates other than those\n  necessary to fulfil the new packages\' dependencies.\n\n  Rationale: Keeping upgrades and other packages changes apart is useful when reading/bisecting\n  history. It also makes code review easier.\n\n- When there\'s a merge conflict in poetry.lock, first check whether either or both commits contain a\n  package upgrade:\n\n  - If neither of them do, simply `git checkout --ours -- poetry.lock && poetry lock --no-update`.\n  - If one of them does, check out that file (`git checkout --ours -- poetry.lock` or\n    `git checkout --theirs -- poetry.lock`) and run `poetry lock --no-update` to regenerate\n    `poetry.lock` with the current package versions.\n  - If both of them do, manually merge `poetry.lock` and run `poetry lock --no-update`.\n\n  Rationale: This should avoid accidentally down- or upgrading when resolving a merge conflict.\n\n- Update the code coverage minimum in pyproject.toml and the badge above on branches which increase\n  it.\n\n  Rationale: By updating this continuously we avoid missing test regressions in new branches.\n\n### Upgrading Python version\n\nTo minimise the chance of discrepancies between environments it is important to run the same (or as\nclose as possible) version of Python in the development environment, in the pipeline, and in\ndeployed instances. At the moment the available versions are constrained by the following:\n\n- The\n  [Ubuntu packages](https://packages.ubuntu.com/search?keywords=python3&searchon=names&exact=1&suite=all&section=all)\n  used in the [Dockerfile](/linz/geostore/blob/master/geostore/Dockerfile)\n- The [AWS base images](https://docs.aws.amazon.com/lambda/latest/dg/python-image.html) used as\n  [Lambda runtimes](/linz/geostore/blob/master/infrastructure/constructs/lambda_config.py)\n- The [pyenv versions](https://github.com/pyenv/pyenv) used for\n  [local development](/linz/geostore/blob/master/.python-version)\n- The [supported Poetry versions](https://python-poetry.org/docs/#system-requirements) used for all\n  [dependencies](/linz/geostore/blob/master/pyproject.toml)\n\nWhen updating Python versions you have to check that all of the above can be kept at the same minor\nversion, and ideally at the same patch level.\n\n### Running tests\n\nPrerequisites:\n\n- Authenticated to a profile which has access to a deployed Geostore.\n\nTo launch full test suite, run `pytest`.\n\n### Debugging\n\nTo start debugging at a specific line, insert `import ipdb; ipdb.set_trace()`.\n\nTo debug a test run, add `--capture=no` to the `pytest` arguments. You can also automatically start\ndebugging at a test failure point with `--pdb --pdbcls=IPython.terminal.debugger:Pdb`.\n\n### Upgrading CI runner\n\n[`jobs.<job_id>.runs-on`](https://docs.github.com/en/free-pro-team@latest/actions/reference/workflow-syntax-for-github-actions#jobsjob_idruns-on)\nin .github sets the runner type per job. We should make sure all of these use the latest specific\n("ubuntu-YY.MM" as opposed to "ubuntu-latest") Ubuntu LTS version, to make sure the version changes\nonly when we\'re ready for it.\n\n### GitHub Actions cache clearing\n\nTo throw away the current cache (for example in case of a cache corruption), simply change the\n[`CACHE_SEED` repository "secret"](https://github.com/linz/geostore/settings/secrets/actions/CACHE_SEED),\nfor example to the current timestamp (`date +%s`). Subsequent jobs will then ignore the existing\ncache.\n',
    'author': 'Bill M. Nelson',
    'author_email': 'bmnelson@linz.govt.nz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/linz/geostore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from json import dumps

from jsonschema import validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY, SUCCESS_KEY
from ..import_file_batch_job_id_keys import ASSET_JOB_ID_KEY, METADATA_JOB_ID_KEY
from ..step_function import get_tasks_status
from ..step_function_keys import (
    ASSET_UPLOAD_KEY,
    DATASET_ID_KEY,
    IMPORT_DATASET_KEY,
    JOB_STATUS_RUNNING,
    METADATA_UPLOAD_KEY,
    VALIDATION_KEY,
    VERSION_ID_KEY,
)
from ..types import JsonObject

INPUT_KEY = "input"
EXECUTION_ID_KEY = "execution_id"

LOGGER = get_log()


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    LOGGER.debug(dumps({EVENT_KEY: event}))

    validate(
        event,
        {
            "type": "object",
            "properties": {
                DATASET_ID_KEY: {"type": "string"},
                VERSION_ID_KEY: {"type": "string"},
                VALIDATION_KEY: {
                    "type": "object",
                    "properties": {SUCCESS_KEY: {"type": "boolean"}},
                    "required": [SUCCESS_KEY],
                },
                IMPORT_DATASET_KEY: {
                    "type": "object",
                    "properties": {
                        METADATA_JOB_ID_KEY: {"type": "string"},
                        ASSET_JOB_ID_KEY: {"type": "string"},
                    },
                    "required": [METADATA_JOB_ID_KEY, ASSET_JOB_ID_KEY],
                },
            },
            "required": [DATASET_ID_KEY, VERSION_ID_KEY, VALIDATION_KEY, IMPORT_DATASET_KEY],
        },
    )

    raw_import_status = get_tasks_status(
        JOB_STATUS_RUNNING,
        event[DATASET_ID_KEY],
        event[VERSION_ID_KEY],
        event[VALIDATION_KEY][SUCCESS_KEY],
        {
            METADATA_JOB_ID_KEY: event[IMPORT_DATASET_KEY][METADATA_JOB_ID_KEY],
            ASSET_JOB_ID_KEY: event[IMPORT_DATASET_KEY][ASSET_JOB_ID_KEY],
        },
    )
    return {
        key: raw_import_status[key]
        for key in [VALIDATION_KEY, ASSET_UPLOAD_KEY, METADATA_UPLOAD_KEY]
        if key in raw_import_status
    }



import sys
from enum import IntEnum
from http import HTTPStatus
from json import dumps, load
from os import environ
from typing import Callable, Optional, Union

import boto3
from botocore.exceptions import NoCredentialsError, NoRegionError
from typer import Option, Typer, secho
from typer.colors import GREEN, RED, YELLOW

from geostore.datasets.create import TITLE_CHARACTERS
from geostore.environment import ENV_NAME_VARIABLE_NAME, PRODUCTION_ENVIRONMENT_NAME

from .api_keys import MESSAGE_KEY
from .aws_keys import BODY_KEY, HTTP_METHOD_KEY, STATUS_CODE_KEY
from .dataset_keys import DATASET_KEY_SEPARATOR
from .resources import Resource
from .step_function_keys import (
    DATASET_ID_SHORT_KEY,
    DESCRIPTION_KEY,
    EXECUTION_ARN_KEY,
    METADATA_URL_KEY,
    S3_ROLE_ARN_KEY,
    TITLE_KEY,
    VERSION_ID_KEY,
)
from .types import JsonList, JsonObject

DATASET_ID_HELP = "Dataset ID, as printed when running `geostore dataset create`."

HTTP_METHOD_CREATE = "POST"

app = Typer(context_settings=dict(max_content_width=sys.maxsize))
dataset_app = Typer(help="Operate on entire datasets.")
dataset_version_app = Typer(help="Operate on dataset versions.")
app.add_typer(dataset_app, name="dataset")
app.add_typer(dataset_version_app, name="version")

GetOutputFunctionType = Union[Callable[[JsonList], str], Callable[[JsonObject], str]]


class ExitCode(IntEnum):
    SUCCESS = 0
    UNKNOWN = 1
    # Exit code 2 is used by Typer to indicate usage error
    CONFLICT = 3
    NO_CREDENTIALS = 4
    NO_REGION_SETTING = 5


@app.callback()
def main(
    environment_name: Optional[str] = Option(
        None,
        help="Set environment name, such as 'test'."
        f" Overrides the value of ${ENV_NAME_VARIABLE_NAME}."
        f"  [default: {PRODUCTION_ENVIRONMENT_NAME}]",
    )
) -> None:
    if environment_name:
        environ[ENV_NAME_VARIABLE_NAME] = environment_name
    elif ENV_NAME_VARIABLE_NAME not in environ:
        environ[ENV_NAME_VARIABLE_NAME] = PRODUCTION_ENVIRONMENT_NAME


@dataset_app.command(name="create", help="Create a new dataset.")
def dataset_create(
    title: str = Option(..., help=f"Allowed characters: '{TITLE_CHARACTERS}'."),
    description: str = Option(...),
) -> None:
    request_object = {
        HTTP_METHOD_KEY: HTTP_METHOD_CREATE,
        BODY_KEY: {TITLE_KEY: title, DESCRIPTION_KEY: description},
    }

    def get_output(response_body: JsonObject) -> str:
        dataset_id: str = response_body[DATASET_ID_SHORT_KEY]
        return dataset_id

    handle_api_request(
        Resource.DATASETS_ENDPOINT_FUNCTION_NAME.resource_name, request_object, get_output
    )


@dataset_app.command(name="list", help="List datasets.")
def dataset_list(id_: Optional[str] = Option(None, "--id", help=DATASET_ID_HELP)) -> None:
    body = {}
    get_output: GetOutputFunctionType

    if id_ is None:

        def get_list_output(response_body: JsonList) -> str:
            lines = []
            for entry in response_body:
                lines.append(
                    f"{entry[TITLE_KEY]}{DATASET_KEY_SEPARATOR}{entry[DATASET_ID_SHORT_KEY]}"
                )
            return "\n".join(lines)

        get_output = get_list_output

    else:

        def get_single_output(response_body: JsonObject) -> str:
            return f"{response_body['title']}{DATASET_KEY_SEPARATOR}{response_body['id']}"

        body[DATASET_ID_SHORT_KEY] = id_
        get_output = get_single_output

    handle_api_request(
        Resource.DATASETS_ENDPOINT_FUNCTION_NAME.resource_name,
        {HTTP_METHOD_KEY: "GET", BODY_KEY: body},
        get_output,
    )


@dataset_app.command(name="delete", help="Delete a dataset.")
def dataset_delete(id_: str = Option(..., "--id", help=DATASET_ID_HELP)) -> None:
    handle_api_request(
        Resource.DATASETS_ENDPOINT_FUNCTION_NAME.resource_name,
        {HTTP_METHOD_KEY: "DELETE", BODY_KEY: {DATASET_ID_SHORT_KEY: id_}},
        None,
    )


@dataset_version_app.command(name="create", help="Create a dataset version.")
def dataset_version_create(
    dataset_id: str = Option(..., help=DATASET_ID_HELP),
    metadata_url: str = Option(
        ...,
        help="S3 URL to the top level metadata file,"
        " for example 's3://my-bucket/my-dataset/collection.json'.",
    ),
    s3_role_arn: str = Option(
        ...,
        help="ARN of the role which the Geostore should assume to read your dataset,"
        " for example 'arn:aws:iam::1234567890:role/s3-reader'.",
    ),
) -> None:
    def get_output(response_body: JsonObject) -> str:
        return f"{response_body[VERSION_ID_KEY]}\t{response_body[EXECUTION_ARN_KEY]}"

    handle_api_request(
        Resource.DATASET_VERSIONS_ENDPOINT_FUNCTION_NAME.resource_name,
        {
            HTTP_METHOD_KEY: HTTP_METHOD_CREATE,
            BODY_KEY: {
                DATASET_ID_SHORT_KEY: dataset_id,
                METADATA_URL_KEY: metadata_url,
                S3_ROLE_ARN_KEY: s3_role_arn,
            },
        },
        get_output,
    )


def handle_api_request(
    function_name: str, request_object: JsonObject, get_output: Optional[GetOutputFunctionType]
) -> None:
    response_payload = invoke_lambda(function_name, request_object)
    status_code = response_payload[STATUS_CODE_KEY]
    response_body = response_payload[BODY_KEY]

    if status_code in [HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.NO_CONTENT]:
        if get_output is not None:
            output = get_output(response_body)
            secho(output, fg=GREEN)
        sys.exit(ExitCode.SUCCESS)

    if status_code == HTTPStatus.CONFLICT:
        secho(response_body[MESSAGE_KEY], err=True, fg=YELLOW)
        sys.exit(ExitCode.CONFLICT)

    secho(dumps(response_body), err=True, fg=RED)
    sys.exit(ExitCode.UNKNOWN)


def invoke_lambda(function_name: str, request_object: JsonObject) -> JsonObject:
    try:
        client = boto3.client("lambda")
    except NoRegionError:
        secho(
            "Unable to locate region settings. Make sure to log in to AWS first.",
            err=True,
            fg=YELLOW,
        )
        sys.exit(ExitCode.NO_REGION_SETTING)

    request_payload = dumps(request_object).encode()

    try:
        response = client.invoke(FunctionName=function_name, Payload=request_payload)
    except NoCredentialsError:
        secho(
            "Unable to locate credentials. Make sure to log in to AWS first.", err=True, fg=YELLOW
        )
        sys.exit(ExitCode.NO_CREDENTIALS)

    response_payload: JsonObject = load(response["Payload"])
    return response_payload


if __name__ == "__main__":
    app()



"""List all datasets function."""
from http import HTTPStatus

from ..api_responses import success_response
from ..datasets_model import datasets_model_with_meta
from ..models import DATASET_ID_PREFIX
from ..types import JsonObject


def list_datasets() -> JsonObject:
    """GET: List all Datasets."""

    # list all datasets
    datasets_model_class = datasets_model_with_meta()
    datasets = datasets_model_class.scan(
        filter_condition=datasets_model_class.id.startswith(DATASET_ID_PREFIX)
    )

    # return response
    resp_body = []
    for dataset in datasets:
        resp_item = dataset.as_dict()
        resp_body.append(resp_item)

    return success_response(HTTPStatus.OK, resp_body)



#!/usr/bin/env python3
from argparse import ArgumentParser, Namespace

from linz_logger import get_log

from ..models import DB_KEY_SEPARATOR
from ..processing_assets_model import ProcessingAssetType
from ..step_function import get_hash_key
from ..validation_results_model import ValidationResultFactory
from .utils import ChecksumValidator, get_job_offset

LOGGER = get_log()


def parse_arguments() -> Namespace:
    argument_parser = ArgumentParser()
    argument_parser.add_argument("--dataset-id", required=True)
    argument_parser.add_argument("--version-id", required=True)
    argument_parser.add_argument("--first-item", type=int, required=True)
    argument_parser.add_argument("--results-table-name", required=True)
    argument_parser.add_argument("--assets-table-name", required=True)
    argument_parser.add_argument("--s3-role-arn", required=True)
    return argument_parser.parse_args()


def main() -> None:
    arguments = parse_arguments()

    index = arguments.first_item + get_job_offset()
    hash_key = get_hash_key(arguments.dataset_id, arguments.version_id)
    range_key = f"{ProcessingAssetType.DATA.value}{DB_KEY_SEPARATOR}{index}"

    validation_result_factory = ValidationResultFactory(hash_key, arguments.results_table_name)

    checksum_validator = ChecksumValidator(
        arguments.assets_table_name, validation_result_factory, arguments.s3_role_arn, LOGGER
    )

    checksum_validator.validate(hash_key, range_key)


if __name__ == "__main__":
    main()



from typing import Final

JOB_STATUS_FAILED = "FAILED"
JOB_STATUS_RUNNING = "RUNNING"
JOB_STATUS_SUCCEEDED = "SUCCEEDED"

S3_BATCH_STATUS_FAILED: Final = "Failed"
S3_BATCH_STATUS_CANCELLED: Final = "Cancelled"
S3_BATCH_STATUS_COMPLETE: Final = "Complete"

ASSET_UPLOAD_KEY = "asset_upload"
DATASET_ID_KEY = "dataset_id"
DATASET_ID_SHORT_KEY = "id"
DATASET_PREFIX_KEY = "dataset_prefix"
DESCRIPTION_KEY = "description"
ERRORS_KEY = "errors"
ERROR_CHECK_KEY = "check"
ERROR_DETAILS_KEY = "details"
ERROR_RESULT_KEY = "result"
ERROR_URL_KEY = "url"
EXECUTION_ARN_KEY = "execution_arn"
FAILED_TASKS_KEY = "failed_tasks"
FAILURE_REASONS_KEY = "failure_reasons"
IMPORT_DATASET_KEY = "import_dataset"
INPUT_KEY = "input"
METADATA_UPLOAD_KEY = "metadata_upload"
METADATA_URL_KEY = "metadata_url"
NEW_VERSION_S3_LOCATION = "new_version_s3_location"
NOW_KEY = "now"
OUTPUT_KEY = "output"
S3_BATCH_RESPONSE_KEY = "s3_batch_response"
S3_ROLE_ARN_KEY = "s3_role_arn"
STATUS_KEY = "status"
STEP_FUNCTION_KEY = "step_function"
TITLE_KEY = "title"
UPDATE_DATASET_KEY = "update_dataset_catalog"
UPLOAD_STATUS_KEY = "upload_status"
VALIDATION_KEY = "validation"
VERSION_ID_KEY = "version_id"



from json import dumps

from botocore.exceptions import ClientError
from botocore.response import StreamingBody
from jsonschema import ValidationError, validate
from linz_logger import get_log

from ..api_keys import EVENT_KEY
from ..error_response_keys import ERROR_KEY, ERROR_MESSAGE_KEY
from ..parameter_store import ParameterName, get_param
from ..s3 import get_s3_client_for_role
from ..s3_utils import get_bucket_and_key_from_url
from ..step_function import get_hash_key
from ..step_function_keys import DATASET_ID_KEY, METADATA_URL_KEY, S3_ROLE_ARN_KEY, VERSION_ID_KEY
from ..types import JsonObject
from ..validation_results_model import ValidationResultFactory
from .utils import STACDatasetValidator

LOGGER = get_log()


def lambda_handler(event: JsonObject, _context: bytes) -> JsonObject:
    LOGGER.debug(dumps({EVENT_KEY: event}))

    # validate input
    try:
        validate(
            event,
            {
                "type": "object",
                "properties": {
                    DATASET_ID_KEY: {"type": "string"},
                    VERSION_ID_KEY: {"type": "string"},
                    METADATA_URL_KEY: {"type": "string"},
                    S3_ROLE_ARN_KEY: {"type": "string"},
                },
                "required": [DATASET_ID_KEY, METADATA_URL_KEY, S3_ROLE_ARN_KEY, VERSION_ID_KEY],
                "additionalProperties": True,
            },
        )
    except ValidationError as error:
        LOGGER.warning(dumps({ERROR_KEY: error}, default=str))
        return {ERROR_MESSAGE_KEY: error.message}

    try:
        s3_client = get_s3_client_for_role(event[S3_ROLE_ARN_KEY])
    except ClientError as error:
        LOGGER.warning(dumps({ERROR_KEY: error}, default=str))
        return {ERROR_MESSAGE_KEY: str(error)}

    def s3_url_reader(url: str) -> StreamingBody:
        bucket_name, key = get_bucket_and_key_from_url(url)
        response = s3_client.get_object(Bucket=bucket_name, Key=key)
        return response["Body"]

    hash_key = get_hash_key(event[DATASET_ID_KEY], event[VERSION_ID_KEY])

    validation_result_factory = ValidationResultFactory(
        hash_key, get_param(ParameterName.STORAGE_VALIDATION_RESULTS_TABLE_NAME)
    )
    validator = STACDatasetValidator(hash_key, s3_url_reader, validation_result_factory)

    validator.run(event[METADATA_URL_KEY])
    return {}



from enum import Enum


class Check(Enum):
    ASSETS_IN_DATASET = "assets in dataset"
    CHECKSUM = "checksum"
    DUPLICATE_OBJECT_KEY = "duplicate asset name"
    JSON_PARSE = "JSON parse"
    JSON_SCHEMA = "JSON schema"
    STAGING_ACCESS = "staging bucket access"
    NON_S3_URL = "not an s3 url"
